# This Python file uses the following encoding: utf-8
import getpass
import logging
import os
import re
import typing
from functools import partial

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QModelIndex, Qt, QPointF, QPoint, QSize, QEvent
from PySide6.QtGui import QColor, QKeySequence
from PySide6.QtGui import QShortcut, QAction, QBrush, QPixmap, QPainter, QFont, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMenu,
    QMessageBox,
    QLabel,
    QTreeWidgetItem,
    QGraphicsScene)
from PySide6.QtWidgets import QPushButton, QWidget, QTreeWidget, QHBoxLayout, QVBoxLayout, \
    QScrollArea, QTextEdit, QGraphicsView, QFrame, QSlider
from PySide6.QtWidgets import QSizePolicy
from fameio.input.scenario import Attribute
from fameio.input.schema import AttributeType

""" needs to be imported to load the resources """
""" Some IDEs may mark this as unused, but it is needed to load the resources """
# noinspection PyUnresolvedReferences
import famegui.generated.qt_resources_rc
from famegui import models
from famegui.agent_controller import AgentController
from famegui.appworkingdir import AppWorkingDir
from famegui.config.runtime_consts import CONTRACT_RECEIVER_ID, UPDATED_CONTRACT_KEY, OLD_CONTRACT_KEY, \
    CONTRACT_SENDER_ID, CONTRACT_PRODUCT_NAME
from famegui.data_manager.loader import load_db_data
from famegui.database.db_management import (
    get_recently_opened_projects,
    manage_path_for_db)
from famegui.database.prebuild_queries.general import get_stored_shortcuts
from famegui.dialogs.ask_for_agent_deletion_dialog import DeletionDialog
from famegui.dialogs.dialog_new_multi_dialog import DialogNewMultiContract
from famegui.dialogs.dialog_newagent import DialogNewAgent
from famegui.dialogs.dialog_newcontract import DialogNewContract
from famegui.dialogs.dialog_scenario_properties import DialogScenarioProperties
from famegui.dialogs.edit_agent_dialog import EditAgentDialog
from famegui.dialogs.edit_contract_dialog import EditContractDialog
from famegui.dialogs.famegui_config_presets_dialog import EditFameDbValuesDialog
from famegui.dialogs.multi_edit_contract_dialog import MultiEditContractDialog
from famegui.famegui_widgets.TimeSeriesPlotCanvasWidget import TimeSeriesPlotCanvas
from famegui.generated.ui_mainwindow import Ui_MainWindow
from famegui.maincontroller import MainController
from famegui.model_utils.contract_helper import get_contract_from_fields
from famegui.models import Agent, Contract
from famegui.path_resolver import FameGuiPathResolver
from famegui.scenario_canvas_elements.agent_canvas_item import AgentGraphItem
from famegui.scenario_canvas_elements.contract_canvas_item import ContractGraphItem
from famegui.scenario_graph_view import ScenarioGraphView
from famegui.ui.fame_ui_elements import QFameBoldLabel
from famegui.ui.main_graphics_widget import MainGraphicsWidget
from famegui.ui.quick_modals import gen_quick_warning_modal
from famegui.ui.ui_utils import GUIConsoleHandler
from famegui.utils import get_product_name_from_contract
from famegui_runtime_helper.dict_hash_helper import hash_dict


class PropertyTreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(
            self,
            parent: QtWidgets.QTreeWidget,
            attr_name: str,
            attr_value: Attribute,
    ):
        super().__init__(parent, [attr_name, str(attr_value.value)])
        self.setFlags(self.flags() | QtCore.Qt.ItemIsEditable)

    def setData(self, column, role, value):
        """Override QTreeWidgetItem.setData()"""
        if role == QtCore.Qt.EditRole:
            # review: Can we remove this code? It seems to be unused at the moment and not necessary.
            # logging.info("new value: {}".format(value))
            pass

        QtWidgets.QTreeWidgetItem.setData(self, column, role, value)


class MainWindow(QMainWindow):
    """Application entry point"""

    def __init__(self, working_dir: AppWorkingDir, ui_file_loader):
        super().__init__()

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self._ui_file_loader = ui_file_loader
        self._working_dir = working_dir
        self._path_resolver = FameGuiPathResolver(self._working_dir)
        self._tree_items_for_agent_types = {}
        self._controller = MainController(self._working_dir)

        # init
        self._init_ui()
        self._connect_actions()
        self._connect_slots()
        self._on_project_closed()

        self._graph_view.installEventFilter(self)

        load_db_data()

        self._setup_shortcuts()

    def on_ctrl_x_activated(self):
        modifiers = QApplication.keyboardModifiers()

        # Check if Shift modifier was held down
        if modifiers == QtCore.Qt.KeyboardModifier.ShiftModifier:
            logging.info("Shift key was pressed along with the shortcut!")
        else:
            logging.info("Shortcut activated without Shift.")

        self.rearrange_layout()

    def handle_redo(self):
        """Performs redo OP if it's triggered by the toolbar or a shortcut"""
        logging.info("Performing redo ...")
        self._controller.perform_redo()

    def handle_undo(self):
        """Performs undo OP if it's triggered by the toolbar or a shortcut"""

        if not self._controller.is_open:
            return

        self._controller.perform_undo()
        self._graph_view.update_all_contracts()

    def set_item_background_color_till_depth(self, item, color, current_depth=0):
        """
        Set the background color for QTreeWidgetItems explicit agents (agent nodes)
        """
        if current_depth <= 1:  # Apply color only if depth is 0 or 1
            for col in range(item.columnCount()):
                item.setBackground(col, color)

        if current_depth < 1:  # Recursively handle children if depth is less than 1
            for i in range(item.childCount()):
                self.set_item_background_color_till_depth(item.child(i), color, current_depth + 1)

    def handle_agent_color_change(self):
        """Apply newly assigned agent type color on all associated ui elements"""

        agent_ids_to_type = self._controller.get_agent_ids_to_type()

        ui_type_to_color_mapping, agent_type_names_from_db = (
            self._controller.get_agent_type_to_color_with_type_listing())

        for agent_graph_item in self._graph_view.get_agent_items()[:]:
            agent_graph_item: AgentGraphItem
            color = ui_type_to_color_mapping[agent_ids_to_type[agent_graph_item.agent_id]]

            agent_graph_item.set_bg_color(color)
            agent_graph_item.update()

        for i in range(self.ui.treeProjectStructure.topLevelItemCount()):
            tree_widget_item = self.ui.treeProjectStructure.topLevelItem(i)
            agent_type = tree_widget_item.text(0)

            color = QColor(ui_type_to_color_mapping[agent_type])

            self.set_item_background_color_till_depth(tree_widget_item, color)

    def handle_db_load_shortcut(self):

        load_db_data()

    def _init_ui(self):
        self._setup_tool_bar_nav_items()
        logging.debug("initializing main window UI")
        self.setWindowIcon(QtGui.QIcon(":/icons/nodes-128px.png"))

        # create and attach the scene
        self.fame_graphicsView = MainGraphicsWidget()
        self.fame_graphicsView.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.TextAntialiasing)
        self.fame_graphicsView.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.BoundingRectViewportUpdate)

        self._graph_view = ScenarioGraphView(
            self._on_agent_dragged,
            self)

        self._graph_view.setSceneRect(self._controller.compute_scene_rect())
        self.fame_graphicsView.setScene(self._graph_view)
        # important: set the index method to NoIndex, otherwise the scene will not be updated correctly and crash
        self._graph_view.setItemIndexMethod(QGraphicsScene.ItemIndexMethod.NoIndex)

        # customize main window
        self.ui.labelUserName.setText(getpass.getuser())
        self.fame_graphicsView.setBackgroundBrush(QtGui.QColor(194, 194, 194))

        self.fame_graphicsView.unselect_contracts.connect(self.clear_contract_highlights)
        self.ui.consoleLogTextEdit.setStyleSheet("background: #2D2D2D;")
        self.ui.consoleLogTextEdit.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.ui.consoleLogTextEdit.setMinimumSize(300, 100)
        self.fame_graphicsView.setMinimumSize(400, 400)
        self.fame_graphicsView.setBaseSize(600, 600)

        self.setWindowTitle(QtWidgets.QApplication.instance().applicationDisplayName())

        # Zoom-related setup
        self.fame_graphicsView.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.fame_graphicsView.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.fame_graphicsView.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.fame_graphicsView.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.fame_graphicsView.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)

        # May enhance performance with hardware acceleration
        # However Anti-Aliasing is not kicking in on far distances
        # self.fame_graphicsView.setViewport(QOpenGLWidget())

        self.fame_graphicsView.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.fame_graphicsView.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)

        self.ui.centralwidget.setStretchFactor(0, 1)  # Top frame
        self.ui.centralwidget.setStretchFactor(1, 4)  # Graphics view
        self.ui.centralwidget.setStretchFactor(2, 2)  # Console

        bottom_box = QWidget()

        logger = logging.getLogger()
        clearButton = QPushButton("Clear")
        self.ui.verticalLayoutBottomDeep.addWidget(clearButton)
        label_main = QTextEdit()
        bottom_box_layout_v = QVBoxLayout()
        self.sliderZoomFactor = QSlider()

        self.labelZoomFactor = QLabel()
        font2 = QFont()
        font2.setPointSize(10)
        self.labelZoomFactor.setFont(font2)

        zoom_lvl_layout = QHBoxLayout()
        zoom_lvl_layout.addWidget(self.labelZoomFactor)
        zoom_lvl_layout.addWidget(self.sliderZoomFactor)

        slider_control_panel_widget = QWidget()
        slider_control_panel_widget.setLayout(zoom_lvl_layout)

        bottom_box_layout_v.addWidget(slider_control_panel_widget)
        bottom_box_layout_v.addWidget(label_main)

        bottom_box.setLayout(bottom_box_layout_v)

        logger.addHandler(GUIConsoleHandler(label_main, clearButton))

        for i in reversed(range(self.ui.centralwidget.count())):
            widget = self.ui.centralwidget.widget(i)
            if widget is not None:
                widget.setParent(None)

        top_box = QWidget()
        top_box.setStyleSheet("background-color: lightblue;")

        top_box_layout = QVBoxLayout()
        graphics_frame = QFrame()

        top_box_layout.addWidget(self.fame_graphicsView)

        # top_box.setLayout(top_box_layout)
        graphics_frame.setLayout(top_box_layout)

        self.ui.centralwidget.addWidget(graphics_frame)
        self.ui.centralwidget.addWidget(bottom_box)

        self.sliderZoomFactor.setObjectName(u"sliderZoomFactor")
        self.sliderZoomFactor.setMaximumSize(QSize(200, 1000))
        self.sliderZoomFactor.setMinimum(10)
        self.sliderZoomFactor.setMaximum(200)
        self.sliderZoomFactor.setSingleStep(10)
        self.sliderZoomFactor.setValue(100)
        self.sliderZoomFactor.setOrientation(Qt.Orientation.Horizontal)

        # allowed zoom range
        self.sliderZoomFactor.setRange(10, 1000)
        # status bar
        self._status_label_icon = QtWidgets.QLabel()
        self.statusBar().addWidget(self._status_label_icon)

        # project structure tree view
        self.ui.treeProjectStructure.all_agents_deletion_requested.connect(
            self._del_all_agents_of_type
        )

        self.ui.treeProjectStructure.setColumnCount(1)
        self.ui.treeProjectStructure.setHeaderLabels(["Agents"])
        # attributes tree view
        self.ui.treeAttributes.setSelectionMode(
            QtWidgets.QAbstractItemView.SelectionMode.SingleSelection
        )

        self.ui.treeAttributes.setRootIsDecorated(False)
        self.ui.treeAttributes.setColumnCount(2)
        self.ui.treeAttributes.setHeaderLabels(["Attribute", "Value"])
        self.ui.treeAttributes.setColumnWidth(0, 140)
        self.ui.treeAttributes.setAlternatingRowColors(True)
        self.ui.treeAttributes.setEditTriggers(
            QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers
        )

        # console
        self.ui.consoleLogTextEdit.setMinimumHeight(70)

        # init global logger
        logger = logging.getLogger()
        self.ui.treeProjectStructure.clicked.connect(self.on_clicked_tree)

        # recently opened projects
        recently_opened_projects = get_recently_opened_projects()

        if len(recently_opened_projects) == 0:
            return

        another_menu = QMenu("Recently opened projects", self)

        for project in recently_opened_projects:
            action = QAction(project.path, self)
            action.triggered.connect(partial(self.open_project, project.path))
            another_menu.addAction(action)

        self.ui.menuok.addMenu(another_menu)

    def _setup_tool_bar_nav_items(self):
        self._undo_action = QAction(QIcon(), "Perform Undo", self)

        self._redo_action = QAction(QIcon(), "Perform Redo", self)

        self._undo_action.triggered.connect(self.handle_undo)
        self._redo_action.triggered.connect(self.handle_redo)

        self.ui.menu_Edit.addAction(self._undo_action)
        self.ui.menu_Edit.addAction(self._redo_action)
        self.ui.menu_Edit.addAction(self.ui.actionConfigMenu)

    def _on_agent_dragged(self, agent_id, agent_x, agent_y, agent_x_t, agent_y_t):

        self._controller.agent_model_was_modified(agent_id, agent_x, agent_y)

    def open_project(self, file_path):

        if not self._confirm_current_project_can_be_closed():
            return
        if file_path != "":
            manage_path_for_db(file_path)

            self.load_scenario_file(file_path)

    def _del_all_agents_of_type(self, agent_type_name: str):
        selected_agent_type_list = self._controller.get_all_agents_by_type_name(
            agent_type_name
        )
        DeletionDialog(
            selected_agent_type_list,
            scenario=self._controller.scenario,
            working_dir=self._working_dir,
            parent_widget=self.ui.widget,
            main_controller=self._controller,
        )

    def _connect_actions(self):
        logging.debug("connecting main window actions")
        # new
        self.ui.actionNewProject.triggered.connect(self.new_project)
        # open
        self.ui.actionOpenProject.triggered.connect(self.show_open_scenario_file_dlg)
        # save (enabled only when a change has been done)
        self.ui.actionSaveProject.triggered.connect(self.save_project)
        # save as
        self.ui.actionSaveProjectAs.triggered.connect(self.save_project_as)
        # close
        self.ui.actionCloseProject.triggered.connect(self.close_project)

        self.ui.actionCloseProject.setVisible(False)
        # generate protobuf
        self.ui.actionMakeRunConfig.triggered.connect(self.make_run_config)
        # exit
        self.ui.actionExit.triggered.connect(self.close)

        self.ui.actionSchemaValidation.triggered.connect(self.revalidate_scenario)

        self.ui.actionGeneralProperties.triggered.connect(
            self._on_edit_scenario_properties
        )

        self.ui.actionConfigMenu.triggered.connect(
            self._on_edit_fame_gui_settings
        )

        # tree
        self.ui.treeProjectStructure.agent_deletion_requested.connect(
            self._on_single_agent_deletion_requested
        )
        self.ui.treeProjectStructure.single_contract_deletion_requested.connect(
            self._on_single_contract_deletion_requested
        )

        self.ui.treeProjectStructure.on_multi_contract_selected.connect(
            self._on_multi_contract_edit_requested
        )

        # self.ui.

    def _setup_shortcuts(self):
        """Setup Global Shortcuts with by user created preferences or defaults"""
        shortcut_data = get_stored_shortcuts()

        for item in shortcut_data:

            if item["shortcut_name"] == "undo":
                self._undo_action.triggered.connect(self.handle_undo)
                self._undo_action.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_redo":
                self._redo_action.triggered.connect(self.handle_redo)
                self._redo_action.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_load_yaml_into_db":
                self.load_from_db_shortcut = QShortcut(QKeySequence(item["shortcut_key"]), self)

                self.load_from_db_shortcut.setContext(Qt.ShortcutContext.ApplicationShortcut)
                self.load_from_db_shortcut.activated.connect(self.handle_db_load_shortcut)

            if item["shortcut_name"] == "rearrange_layout":
                self.ctrl_x_shortcut = QShortcut(QKeySequence(item["shortcut_key"]), self)
                self.ui.actionReArrangeLayout.triggered.connect(self.rearrange_layout)

                self.ctrl_x_shortcut.activated.connect(self.on_ctrl_x_activated)

            if item["shortcut_name"] == "action_new_project":
                self.ui.actionNewProject.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_open_project":
                self.ui.actionOpenProject.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_save_project":
                self.ui.actionSaveProject.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_quit_app":
                self.ui.actionExit.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_database_properties":
                self.ui.actionConfigMenu.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_config_properties":
                self.ui.actionGeneralProperties.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_make_run_config":
                self.ui.actionMakeRunConfig.setShortcut(QKeySequence(item["shortcut_key"]))

            if item["shortcut_name"] == "action_schema_validation":
                self.ui.actionSchemaValidation.setShortcut(QKeySequence(item["shortcut_key"]))

    def _on_multi_contract_edit_requested(self, selected_contracts: list[dict]):
        """opens up a multi edit  dialog for several selected contracts"""

        selected_contract_models = [get_contract_from_fields(self._controller.scenario,
                                                             contract_data[CONTRACT_PRODUCT_NAME],
                                                             contract_data[CONTRACT_RECEIVER_ID],
                                                             contract_data[CONTRACT_SENDER_ID])
                                    for contract_data in selected_contracts]

        MultiEditContractDialog(selected_contract_models, self._controller.scenario,
                                self._controller,
                                on_contract_edited_slot=self._on_contract_edited)

    def _on_single_agent_deletion_requested(self, agent_id: int):
        """opens up a deletion dialog for a single agent"""
        DeletionDialog(
            [self._controller.agent_from_id(agent_id)],
            scenario=self._controller.scenario,
            working_dir=self._working_dir,
            parent_widget=self.ui.widget,
            main_controller=self._controller,
        )

    def _get_contract_tree_items(
            self, sender_agent: Agent, receiver_agent: Agent, product_name: str
    ) -> typing.List[QTreeWidgetItem]:
        """returns a list of tree leaf items that represent the contract between the two agents"""

        contract_ids = {sender_agent.id, receiver_agent.id}
        contract_tree_items = []
        agent_type_names = {sender_agent.type_name, receiver_agent.type_name}
        for i in range(self.ui.treeProjectStructure.topLevelItemCount()):
            top_lvl_item = self.ui.treeProjectStructure.topLevelItem(i)
            agent_type = top_lvl_item.text(0)

            if agent_type not in agent_type_names:
                continue
            for agent_child_idx in range(top_lvl_item.childCount()):
                selected_node = top_lvl_item.child(agent_child_idx)
                selected_node_id = int(
                    top_lvl_item.child(agent_child_idx).data(0, Qt.ItemDataRole.UserRole)
                )

                if selected_node_id not in contract_ids:
                    continue
                for contract_child_idx in range(selected_node.childCount()):
                    selected_contract_leaf = selected_node.child(contract_child_idx)
                    contract_description = selected_contract_leaf.text(0)
                    if product_name not in contract_description:
                        continue
                    if (
                            str(sender_agent.id) not in contract_description
                            and str(receiver_agent.id) not in contract_description
                    ):
                        continue
                    contract_tree_items.append(selected_contract_leaf)

        return contract_tree_items

    def _remove_contract_tree_items(
            self, sender_agent: Agent, receiver_agent: Agent, product_name: str
    ):
        """removes the contract leaf items between the two agents from the tree"""
        contract_ids = {sender_agent.id, receiver_agent.id}
        contract_leafs_to_remove_idx = []
        contract_to_removes = []

        agent_type_names = {sender_agent.type_name, receiver_agent.type_name}
        for i in range(self.ui.treeProjectStructure.topLevelItemCount()):
            top_lvl_item = self.ui.treeProjectStructure.topLevelItem(i)
            agent_type = top_lvl_item.text(0)

            if agent_type not in agent_type_names:
                continue
            for agent_child_idx in range(top_lvl_item.childCount()):
                selected_node = top_lvl_item.child(agent_child_idx)
                selected_node_id = int(
                    top_lvl_item.child(agent_child_idx).data(0, Qt.ItemDataRole.UserRole)
                )

                if selected_node_id not in contract_ids:
                    continue
                contract_children = [
                    selected_node.child(contract_child_idx)
                    for contract_child_idx in range(selected_node.childCount())
                ]
                for idx, selected_contract_leaf in enumerate(contract_children[:]):
                    contract_description = selected_contract_leaf.text(0)
                    if product_name not in contract_description:
                        continue
                    if (
                            str(sender_agent.id) not in contract_description
                            and str(receiver_agent.id) not in contract_description
                    ):
                        continue

                    contract_leafs_to_remove_idx.append(idx)

                for contract_leaf_to_remove in contract_leafs_to_remove_idx:
                    contract_to_removes.append(
                        {
                            "sender_id": sender_agent.id,
                            "receiver_id": receiver_agent.id,
                            "product_name": get_product_name_from_contract(
                                contract_description
                            ),
                        }
                    )

                    selected_node.removeChild(
                        selected_node.child(contract_leaf_to_remove)
                    )

    def _remove_contract_tree_items_using_contract(self, contract: Contract):
        """removes the contract leaf items between the two agents from the tree using a contract object"""

        self._remove_contract_tree_items(
            self._controller.get_agent_ctrl(contract.sender_id).model,
            self._controller.get_agent_ctrl(contract.receiver_id).model,
            contract.product_name,
        )

    def _on_single_contract_deletion_requested(
            self, sender_agent_id: int, receiver_agent_id: int, product_name: str
    ):
        """opens up a deletion dialog for a single contract"""
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Warning)
        msg_box.setText(
            f"Do you want to delete the contract with the sender: {sender_agent_id}\n.And the receiver: {receiver_agent_id}\nwith the product name:  ({product_name}) ?"
        )
        msg_box.setWindowTitle("Confirmation")
        msg_box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)

        msg_box.setDefaultButton(QMessageBox.No)

        response = msg_box.exec()

        alert_badge = QLabel()
        alert_badge.setText("Alert!")
        alert_badge.setStyleSheet(
            "background-color: red;"
            "color: white;"
            "border-radius: 8px;"
            "padding: 4px 8px;"
        )

        if response == QMessageBox.Yes:
            self._controller.remove_contract(
                sender_agent_id, receiver_agent_id, product_name
            )
            self.ui.treeProjectStructure.blockSignals(True)
            self._graph_view.blockSignals(True)

            self._controller.set_unsaved_changes(True)

            self._remove_contract_tree_items(
                self._controller.get_agent_ctrl(sender_agent_id).model,
                self._controller.get_agent_ctrl(receiver_agent_id).model,
                product_name,
            )

            self._graph_view.remove_contract(sender_agent_id, receiver_agent_id)

            self.ui.treeProjectStructure.blockSignals(False)
            self._graph_view.blockSignals(False)

    def revalidate_scenario(self):
        if not self._controller.is_open:
            return
        schema_valid = self._controller.revalidate_scenario()

        if schema_valid:
            gen_quick_warning_modal(
                self.parentWidget(),
                "Schema Validation",
                "The scenario is valid according to the schema",
            )

    # Contract UI Utils
    def rearrange_layout(self):
        """adjust the layout of the graph view in circular orbital mode"""
        if not self._controller.is_open:
            return
        self._controller.rearrange_layout()
        for item in self.fame_graphicsView.items():
            if issubclass(type(item), ContractGraphItem):
                item: ContractGraphItem
                item.adjust()

    def clear_contract_highlights(self):
        """Resets the highlight mode of all contract items in the graph view"""
        for item in self._graph_view.get_contract_items():
            if item.is_in_highlight_mode():
                item.set_highlight_mode(False)
                item.adjust()

    def _connect_slots(self):
        """Event handling for the main window"""
        logging.debug("initializing main window slots")

        agent_graph_items = self._graph_view

        self.sliderZoomFactor.valueChanged.connect(self._on_zoom_value_changed)

        self._graph_view.contract_creation_requested.connect(
            self._on_contract_creation_requested
        )

        self._graph_view.agent_creation_requested.connect(
            self._on_agent_creation_requested
        )

        self.fame_graphicsView.agent_deletion_requested.connect(
            self._on_agent_deletion_requested
        )

        self._graph_view.agent_edition_requested.connect(
            self._on_agent_edition_requested
        )

        self._graph_view.agent_multi_edition.connect(
            self._on_agent_multi_edition
        )

        self.ui.treeProjectStructure.currentItemChanged.connect(
            self._on_tree_view_current_item_changed
        )

        self.ui.treeProjectStructure.itemDoubleClicked.connect(
            self._on_tree_view_current_item_double_clicked
        )

        self.ui.lineFilterPattern.textChanged.connect(self._filter_pattern_changed)
        self._graph_view.zoom_in_requested.connect(
            lambda: self.sliderZoomFactor.setValue(
                self.sliderZoomFactor.value() + 10
            )
        )
        self._graph_view.released_multi_selection_mode.connect(
            self.on_multi_agent_select
        )

        self._graph_view.released_multi_selection_mode_no_valid.connect(
            self.clear_graph_from_multi_selection
        )
        self._graph_view.zoom_out_requested.connect(
            lambda: self.sliderZoomFactor.setValue(
                self.sliderZoomFactor.value() - 10
            )
        )

        self._controller.project_properties.changed.connect(
            self._on_scenario_status_changed
        )
        self._controller.agent_added.connect(self._on_agent_added)
        self._controller.contract_added.connect(self._on_contract_added)
        self._controller.contract_edited.connect(self._on_contract_edited)
        self._controller.contracts_deleted.connect(self._on_contract_removed)
        self._controller.agents_deleted.connect(self._on_agents_deleted)
        self._controller.selected_agent_changed.connect(self._on_selected_agent_changed)

        self._graph_view.selected_agent_changed.connect(
            self._controller.set_selected_agent_id)

        # undo

    def _on_agents_deleted(self, agent_ids: typing.List[int]):
        """initiates removal of agents from the graph view and the tree view"""
        self._graph_view.clearSelection(True)
        self.clear_contract_highlights()
        self.ui.treeAttributes.clear()
        self.remove_all_except_tree()

        self.ui.treeProjectStructure.blockSignals(True)

        for agent_id in agent_ids:
            agent_ctrl = self._controller.get_agent_ctrl(agent_id)
            self.remove_agent_tree_item(agent_ctrl)

        self._graph_view.remove_agents(agent_ids)

        self.ui.treeProjectStructure.blockSignals(False)

    def _on_agent_multi_edition(self, agent_list: list[str]):

        if not self._controller.check_if_all_agents_have_same_type(agent_list):
            gen_quick_warning_modal(
                self.parentWidget(),
                "Agent Type Mismatch",
                "All agents must have the same type",
            )
            return

        EditAgentDialog(
            agents=[self._controller.agent_from_id(agent_id).model for agent_id in agent_list],
            scenario=self._controller.scenario,
            working_dir=self._working_dir,
            main_ctrl=self._controller,
            ui_loader=self._ui_file_loader,
        )

    def _on_agent_edition_requested(self, agent_id: int):
        """initiates the agent edition dialog"""

        EditAgentDialog(
            agents=[self._controller.agent_from_id(agent_id).model],
            scenario=self._controller.scenario,
            working_dir=self._working_dir,
            main_ctrl=self._controller,
            ui_loader=self._ui_file_loader,
        )

    def _on_agent_deletion_requested(self, q_point: QPoint):
        """initiates the agent deletion dialog"""
        map_point: QPointF = self.fame_graphicsView.mapToScene(q_point)
        agent_id = self._graph_view.get_agent_id(map_point)
        if not agent_id:
            gen_quick_warning_modal(
                self.parentWidget(),
                "No Agent Selected",
                "Please select at least one agent\n"
                "or navigate your mouse to an agent",
            )
            return
        selected_agent = self._controller.agent_from_id(agent_id)
        DeletionDialog(
            [selected_agent],
            scenario=self._controller.scenario,
            working_dir=self._working_dir,
            parent_widget=self.ui.widget,
            main_controller=self._controller,
        )

    def _on_zoom_value_changed(self):
        zoom_factor = self.sliderZoomFactor.value()
        assert zoom_factor > 0
        scale_factor = zoom_factor * 0.001
        self.fame_graphicsView.setTransform(
            QtGui.QTransform.fromScale(scale_factor, scale_factor)
        )
        self.labelZoomFactor.setText("{} %".format(zoom_factor))

    def clear_graph_from_multi_selection(self):
        """Clears the graph view from multi selection mode by clearing the selection and the highlights"""
        self._graph_view.clearSelection(True)
        self.clear_contract_highlights()

    def _get_agent_list_from_ids(
            self, agent_id_list: typing.List[int]
    ) -> typing.List[Agent]:
        """Fetches the agent object list from the agent id list"""
        agent_list = []
        for agent_id in agent_id_list:
            agent_list.append(self._controller.agent_from_id(agent_id))
        return agent_list

    def on_multi_agent_select(self, sender_list: list, receiver_list: list):
        """Initiates the multi contract creation dialog"""

        self.clear_graph_from_multi_selection()
        sender_agent_list = self._get_agent_list_from_ids(sender_list)
        receiver_agent_list = self._get_agent_list_from_ids(receiver_list)

        dlg = DialogNewMultiContract(
            receiver_agent_list,
            sender_agent_list,
            self._controller.scenario.schema,
            self,
        )

        if dlg.exec_() != 0:
            for contract in dlg.make_new_contracts():
                self._controller.add_new_contract(contract)

    def _on_tree_view_current_item_double_clicked(self):
        """initiates a matching edition dialog"""
        assert self._controller.is_open
        tree_item = self.ui.treeProjectStructure.currentItem()

        if tree_item is not None:
            # note: the given id value can be None
            selected_agent_id = tree_item.data(0, QtCore.Qt.ItemDataRole.UserRole)

            if selected_agent_id:
                agent_ctrl_list = [self._controller.agent_from_id(selected_agent_id)]
                agent_model_list = [agent_ctrl.model for agent_ctrl in agent_ctrl_list]

                EditAgentDialog(
                    agents=agent_model_list,
                    scenario=self._controller.scenario,
                    working_dir=self._working_dir,

                    main_ctrl=self._controller,
                    ui_loader=self._ui_file_loader,

                )

                return
            agent_type = tree_item.text(0)
            if agent_type:
                selected_agents = self._controller.get_all_agents_by_type_name(
                    agent_type
                )
                if len(selected_agents) != 0:
                    EditAgentDialog(
                        agents=selected_agents,
                        scenario=self._controller.scenario,
                        working_dir=self._working_dir,
                        main_ctrl=self._controller,
                        ui_loader=self._ui_file_loader,
                    )
                    return

            # single contract selection
            parent_agent_item = tree_item.parent()
            selected_con_agent_id = tree_item.data(2, QtCore.Qt.ItemDataRole.UserRole)
            parent_agent_item_id = parent_agent_item.data(0, QtCore.Qt.ItemDataRole.UserRole)
            product_name = get_product_name_from_contract(
                tree_item.data(1, QtCore.Qt.ItemDataRole.UserRole)
            )

            if tree_item.data(1, QtCore.Qt.ItemDataRole.UserRole) == "receiver":
                edit_contract_dlg = EditContractDialog(
                    self._controller.get_contract(
                        selected_con_agent_id, parent_agent_item_id, product_name
                    ),
                    self._controller.scenario,
                    self._controller,
                    on_contract_edited_slot=self._on_contract_edited,
                )
                edit_contract_dlg.on_contract_edited.connect(self._on_contract_edited)

                return

            EditContractDialog(
                self._controller.get_contract(
                    parent_agent_item_id, selected_agent_id, product_name
                ),
                self._controller.scenario,
                self._controller,
                on_contract_edited_slot=self._on_contract_edited,
            )

    def _on_contract_removed(self, agent_id: int):
        """Clears the graph view and the tree view from the given agent id after removing it from the scenario"""
        related_contracts = self._controller.get_all_related_contracts(agent_id)
        self.ui.treeProjectStructure.blockSignals(True)
        self._graph_view.blockSignals(True)

        for related_contract in related_contracts:
            self._graph_view.remove_contract(
                related_contract.sender_id, related_contract.receiver_id
            )
            self._remove_contract_tree_items_using_contract(related_contract)

        self.ui.treeProjectStructure.blockSignals(False)
        self._graph_view.blockSignals(False)
        self._controller.set_unsaved_changes(True)

    def _on_contract_edited(self, contracts: dict):
        """Updates the graph view and the tree view after editing a contract"""
        self._graph_view.clearSelection()
        old_contract_sender_ctrl = self._controller.get_agent_ctrl(contracts[OLD_CONTRACT_KEY][CONTRACT_SENDER_ID])
        old_contract_receiver_ctrl = self._controller.get_agent_ctrl(contracts[OLD_CONTRACT_KEY][CONTRACT_RECEIVER_ID])
        new_contract_sender_ctrl = self._controller.get_agent_ctrl(contracts[UPDATED_CONTRACT_KEY][CONTRACT_SENDER_ID])
        new_contract_receiver_ctrl = self._controller.get_agent_ctrl(
            contracts[UPDATED_CONTRACT_KEY][CONTRACT_RECEIVER_ID])

        self._graph_view.remove_contract(
            old_contract_sender_ctrl.id,
            old_contract_receiver_ctrl.id,
        )

        self._graph_view.add_contract(
            new_contract_sender_ctrl,
            new_contract_receiver_ctrl,
        )

        self._remove_contract_tree_items(
            old_contract_sender_ctrl.model,
            old_contract_receiver_ctrl.model,
            contracts[OLD_CONTRACT_KEY]["productname"]
        )

        self._create_tree_view_contract(
            new_contract_sender_ctrl,
            new_contract_receiver_ctrl,
            Contract.from_dict(contracts[UPDATED_CONTRACT_KEY])
        )

    def _on_tree_view_current_item_changed(self):
        assert self._controller.is_open
        selected_agent_id = None
        tree_item = self.ui.treeProjectStructure.currentItem()
        self.ui.treeProjectStructure.indexFromItem(tree_item, 0)

        if tree_item is not None:
            # note: the given id value can be None
            selected_agent_id = tree_item.data(0, QtCore.Qt.ItemDataRole.UserRole)
        self._controller.set_selected_agent_id(selected_agent_id)

    def _on_agent_creation_requested_accepted(
            self, dlg: DialogNewAgent, x: int, y: int
    ):
        new_agent = dlg.make_new_agent(self._controller.generate_new_agent_id())
        self._controller.add_new_agent(new_agent, x, y)
        self._graph_view.setSceneRect(self._controller.compute_scene_rect())

        dlg.save_data_in_scenario(data=self._controller.scenario, agent_id=new_agent.id)

    def _on_agent_creation_requested(self, x: int, y: int):
        assert self._controller.is_open

        dlg = DialogNewAgent(
            self._controller.schema, self._working_dir, self._controller.scenario
        )

        if dlg._ui.exec_() != 0:
            new_agent = dlg.make_new_agent(self._controller.generate_new_agent_id())

            self._controller.add_new_agent(new_agent, x, y)

            self._graph_view.setSceneRect(self._controller.compute_scene_rect())

    def _on_contract_creation_requested(self, sender_id: int, receiver_id: int):
        sender = self._controller.agent_from_id(sender_id)
        receiver = self._controller.agent_from_id(receiver_id)

        dlg = DialogNewContract(
            sender, receiver, self._controller.scenario.schema, self
        )

        if dlg.exec_() != 0:
            self._controller.add_new_contract(dlg.make_new_contract())

    def highlight_all_contracts_from_agent(self, agent_id):
        """Highlighting all contracts from a given agent in graph view"""
        for graphic_item in self._graph_view.get_contract_items():
            graphic_item: ContractGraphItem
            if graphic_item.sourceNode().agent_id.__eq__(
                    agent_id
            ) or graphic_item.destNode().agent_id.__eq__(agent_id):
                graphic_item.set_highlight_mode(True)
                graphic_item.adjust()

    def adjust_tree_widget_height(self):
        total_height = self.calculate_tree_widget_height()
        self.ui.treeAttributes.setFixedHeight(total_height)
        self.ui.treeAttributes.resizeColumnToContents(0)
        # self.ui.dockWidgetContents.setFixedHeight(total_height + 30)
        self.ui.dockWidgetContents.updateGeometry()

    def calculate_tree_widget_height(self):
        total_height = self.ui.treeAttributes.header().height()  # Start with the header height

        def get_item_height(item):
            return self.ui.treeAttributes.visualItemRect(item).height()

        def calculate_height(item):
            height = get_item_height(item)
            if item.isExpanded():
                for i in range(item.childCount()):
                    height += calculate_height(item.child(i)) + 10
            return height

        for i in range(self.ui.treeAttributes.topLevelItemCount()):
            total_height += calculate_height(self.ui.treeAttributes.topLevelItem(i)) + 10

        return total_height

    def get_widget_by_name(self, name):
        return self.findChild(QWidget, name)

    def remove_all_except_tree(self):
        # Iterate over all items in the layout
        for i in reversed(range(self.ui.verticalLayout_3.count())):
            widget = self.ui.verticalLayout_3.itemAt(i).widget()
            # Check if the widget is a QTreeWidget
            if not isinstance(widget, QTreeWidget):
                # If not, remove it from the layout and delete it
                self.ui.verticalLayout_3.takeAt(i)
                widget.deleteLater()

    class ListManager:
        def __init__(self):
            self.container = QWidget()
            self.layout = QVBoxLayout(self.container)

            self.scroll_area = QScrollArea()
            self.scroll_area.setWidgetResizable(True)
            self.scroll_area.setWidget(self.container)
            self.widgets = []

        def add_item(self, widget):
            self.layout.addWidget(widget)
            self.widgets.append(widget)

        def remove_item(self):
            if self.widgets:
                widget = self.widgets.pop()
                self.layout.removeWidget(widget)
                widget.deleteLater()

        def clear_items(self):
            while self.widgets:
                self.remove_item()

        def get_scroll_area(self):
            return self.scroll_area

    def can_open_file(self, file_path, mode='r'):
        try:
            with open(file_path, mode):
                return True
        except (FileNotFoundError, IOError, PermissionError) as e:
            logging.error(f"Error: {e}")
            return False

    def eventFilter(self, source, event):
        """Catch Key Release to cancel graph selection mode"""

        if event.type() == QEvent.Type.KeyRelease:
            if event.key() == Qt.Key.Key_Shift:
                self._graph_view.clearSelection(True)

                self._graph_view.set_multi_selection_mode(False)

        # Perform any additional actions here

        # Otherwise, call the base class implementation
        return super().eventFilter(source, event)

    def _on_selected_agent_changed(self, agent_ctrl: AgentController):
        agent: AgentGraphItem

        self.clear_contract_highlights()

        if self._graph_view.is_in_multi_selection_mode():
            return

        if agent_ctrl is None:
            # clear selection
            self.ui.treeProjectStructure.clearSelection()
            self._graph_view.clearSelection()
            self.ui.treeAttributes.clear()
            self.remove_all_except_tree()

            return
        # block signals
        self.ui.treeProjectStructure.blockSignals(True)
        self._graph_view.blockSignals(True)

        # update graph view
        if self._graph_view.is_in_multi_selection_mode():
            return
        self._graph_view.clearSelection()
        agent_ctrl.scene_item.setSelected(True)

        # update tree view
        self.ui.treeProjectStructure.setCurrentItem(agent_ctrl.tree_item)

        # update agent view
        self.ui.treeAttributes.clear()
        self.remove_all_except_tree()

        item_type = QtWidgets.QTreeWidgetItem(
            self.ui.treeAttributes, ["Type", agent_ctrl.type_name]
        )

        item_type.setBackground(1, QBrush(agent_ctrl.svg_color))
        QtWidgets.QTreeWidgetItem(self.ui.treeAttributes, ["ID", agent_ctrl.display_id])

        time_series_dict = {}

        for item1, item2 in self._controller.schema.agent_type_from_name(agent_ctrl.type_name).attributes.items():

            if item2.attr_type == AttributeType.TIME_SERIES:
                time_series_dict[item1] = item2.attr_type

        for attr_name, attr_value in agent_ctrl.attributes.items():

            if attr_name in time_series_dict:
                if attr_value.has_value:
                    time_series_dict[attr_name] = attr_value.value

            PropertyTreeItem(self.ui.treeAttributes, attr_name, attr_value).setToolTip(
                0,
                self._controller.get_help_text_for_attr(
                    agent_ctrl.type_name, attr_name
                ),
            )

        total_height = 0

        list_manager = self.ListManager()

        self.ui.dockWidgetContents.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        self.ui.dockWidgetContents.updateGeometry()

        for idx, (key, value) in enumerate(time_series_dict.items()):

            value = str(value)
            if not value.endswith(".csv"):
                continue

            holder = QWidget()
            holder.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
            obj_name = f"object_plot_{key}"
            holder.setObjectName(obj_name)
            holder.setMinimumHeight(350)

            # Set horizontal layout for the holder widget
            h_layout = QVBoxLayout(holder)
            h_layout.setContentsMargins(0, 0, 0, 0)  # Set margins to 0
            h_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignTop)
            holder.setLayout(h_layout)

            # Add the holder widget to the main vertical layout

            merged_path = os.path.join(str(self._working_dir.root_dir), value)
            merged_path = os.path.normpath(merged_path)

            # Assuming TimeSeriesPlotCanvas is a custom widget that plots the time series
            if self.can_open_file(merged_path):
                TimeSeriesPlotCanvas(holder, merged_path, key)
            else:
                QLabel(f"Path can \n not be opened: {merged_path}", holder)

            combo_holder = QWidget()
            v_item_layout = QVBoxLayout(combo_holder)
            v_item_layout.setContentsMargins(0, 0, 0, 0)
            v_item_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
            v_item_layout.addWidget(holder)

            list_manager.add_item(combo_holder)

            if idx == 0:
                total_height += holder.height()

        self.ui.verticalLayout_3.addWidget(list_manager.get_scroll_area())
        list_manager.get_scroll_area()

        self.ui.verticalLayout_3.addWidget(list_manager.get_scroll_area())

        self.ui.verticalLayout_3.setContentsMargins(0, 0, 0, 0)

        self.ui.dockWidgetContents.setStyleSheet("background-color: #A3A3A3;")

        self.ui.verticalLayout_3.setSpacing(10)

        self.ui.verticalLayout_3.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        for graphic_item in self._graph_view.get_contract_items():
            graphic_item: ContractGraphItem
            if graphic_item.sourceNode().agent_id.__eq__(
                    agent_ctrl.id
            ) or graphic_item.destNode().agent_id.__eq__(agent_ctrl.id):
                graphic_item.set_highlight_mode(True)
                graphic_item.adjust()

        self.adjust_tree_widget_height()

        # unblock signals
        self.ui.treeProjectStructure.blockSignals(False)
        self._graph_view.blockSignals(False)

    def _filter_pattern_changed(self):
        pattern = self.ui.lineFilterPattern.text().lower()
        for a in self._controller.agent_list:
            hide = a.type_name.lower().find(pattern) == -1
            a.tree_item.setHidden(hide)

    def _tree_item_parent_for_agent(self, agent_ctrl) -> QtWidgets.QTreeWidgetItem:
        # return existing if it already exists
        if agent_ctrl.type_name in self._tree_items_for_agent_types:
            return self._tree_items_for_agent_types[agent_ctrl.type_name]
        item = QtWidgets.QTreeWidgetItem(
            self.ui.treeProjectStructure, [agent_ctrl.type_name]
        )

        item.setExpanded(True)
        item.setBackground(0, QBrush(agent_ctrl.svg_color))
        self._tree_items_for_agent_types[agent_ctrl.type_name] = item

        return item

    def remove_agent_tree_item(self, agent_ctrl: AgentController):
        """removes second stage agents from the tree view + top lvl items if there are no more children left"""
        parent_item = self._tree_item_parent_for_agent(agent_ctrl)
        parent_item.removeChild(agent_ctrl.tree_item)
        child_count = parent_item.childCount()

        if child_count != 0:
            return
        for idx in range(0, self.ui.treeProjectStructure.topLevelItemCount()):
            if self.ui.treeProjectStructure.topLevelItem(idx) == parent_item:
                self.ui.treeProjectStructure.takeTopLevelItem(idx)
                del self._tree_items_for_agent_types[agent_ctrl.type_name]

        self.ui.treeProjectStructure.blockSignals(True)
        self._graph_view.blockSignals(True)
        self.ui.treeProjectStructure.blockSignals(False)
        self._graph_view.blockSignals(False)

    def _create_agent_tree_item(self, agent_ctrl: AgentController):
        parent_item = self._tree_item_parent_for_agent(agent_ctrl)
        # create tree item
        item = QtWidgets.QTreeWidgetItem(parent_item, [agent_ctrl.display_id])
        item.setBackground(0, QBrush(agent_ctrl.svg_color))

        item.setData(0, QtCore.Qt.ItemDataRole.UserRole, agent_ctrl.id)
        item.setToolTip(0, agent_ctrl.tooltip_text)
        # add item

        agent_ctrl.tree_item = item
        self.ui.treeProjectStructure.addTopLevelItem(item)

    def _on_agent_added(self, agent_ctrl: AgentController):
        logging.debug("agent_added: {}".format(agent_ctrl.display_id))
        self._graph_view.add_agent(agent_ctrl)
        self._create_agent_tree_item(agent_ctrl)

    def highlight_contract(
            self, sender_id, receiver_id, single_highlight_mode=False
    ):
        """Highlights the contract in the graph view between selected agents"""

        for graphic_item in self._graph_view.get_contract_items():
            graphic_item: ContractGraphItem
            if graphic_item.sourceNode().agent_id.__eq__(
                    sender_id
            ) and graphic_item.destNode().agent_id.__eq__(receiver_id):
                graphic_item.set_highlight_mode(True)
                if single_highlight_mode:
                    graphic_item.set_single_highlight_mode(True)
                graphic_item.adjust()
                return
            elif graphic_item.sourceNode().agent_id.__eq__(
                    receiver_id
            ) and graphic_item.destNode().agent_id.__eq__(sender_id):
                graphic_item.set_highlight_mode(True)
                if single_highlight_mode:
                    graphic_item.set_single_highlight_mode(True)
                graphic_item.adjust()
                return

    def on_clicked_tree(self, pos: QModelIndex):
        """Handles the click event on the agent tree view.
        Process all types of tree items"""

        pos.parent().data(QtCore.Qt.ItemDataRole.UserRole)
        parent_agent_id = pos.parent().data(
            QtCore.Qt.ItemDataRole.UserRole
        )  # get Id from selected agent level

        if parent_agent_id is None:
            return
        results = re.findall(
            r"\d+", str(pos.data(0))
        )  # extract Id of partner agent from the 1 to 1 contract
        if len(results) == 0:  # prevent mismatching
            return
        agent_id = results[0]

        if parent_agent_id is None:
            return

        self.highlight_contract(int(parent_agent_id), int(agent_id), True)

    def _create_tree_view_contract(
            self,
            sender: AgentController,
            receiver: AgentController,
            contract: models.Contract,
    ):
        sender_tree_item = QtWidgets.QTreeWidgetItem(
            sender.tree_item,
            ["{} ({})".format(receiver.display_id, contract.product_name)],
        )

        sender_tree_item.setIcon(0, QtGui.QIcon(":/icons/16px-login.png"))
        sender_tree_item.setData(1, QtCore.Qt.ItemDataRole.UserRole, "sender")
        sender_tree_item.setData(2, QtCore.Qt.ItemDataRole.UserRole, receiver.id)
        sender_tree_item.setBackground(0, QColor(0, 255, 0, 100))  # green color sender

        receiver_tree_item = QtWidgets.QTreeWidgetItem(
            receiver.tree_item,
            [" ({}) {}".format(contract.product_name, sender.display_id)],
        )

        receiver_tree_item.setBackground(
            0, QColor(255, 0, 0, 100)
        )  # red color receiver

        receiver_tree_item.setData(1, QtCore.Qt.ItemDataRole.UserRole, "receiver")
        receiver_tree_item.setData(2, QtCore.Qt.ItemDataRole.UserRole, sender.id)
        receiver_tree_item.setIcon(0, QtGui.QIcon(":/icons/16px-logout.png"))

    def _on_contract_added(
            self,
            sender: AgentController,
            receiver: AgentController,
            contract: models.Contract,
    ):
        """Handles the event when a new contract is added to the scenario by creating UI elements"""
        # update scene graph
        self._graph_view.add_contract(sender, receiver)
        # update tree view
        self._create_tree_view_contract(sender, receiver, contract)

    def _confirm_current_project_can_be_closed(self) -> bool:
        if self._controller.has_unsaved_changes:
            choice = QtWidgets.QMessageBox.warning(
                self,
                self.tr("Modifications will be lost"),
                self.tr(
                    "Modifications done to the current scenario have not been saved!\n\nWhat do you want to do?"
                ),
                QtWidgets.QMessageBox.StandardButtons(
                    QtWidgets.QMessageBox.Save
                    | QtWidgets.QMessageBox.Discard
                    | QtWidgets.QMessageBox.Cancel
                ),
                QtWidgets.QMessageBox.Cancel,
            )
            if choice == QtWidgets.QMessageBox.Save:
                return self.save_project()
            elif choice == QtWidgets.QMessageBox.Discard:
                return True
            else:
                return False
        return True

    def _on_project_closed(self):
        self._graph_view.clear()
        # reset zoom
        self.sliderZoomFactor.setValue(50)
        # reset attributes
        self._tree_items_for_agent_types = {}
        # reset scenario
        self._controller.reset()
        # reset ui
        self.ui.treeProjectStructure.clear()
        self.ui.treeAttributes.clear()
        self.remove_all_except_tree()
        self.ui.lineFilterPattern.clear()
        self.ui.labelProjectName.clear()

    def display_error_msg(self, msg: str) -> None:
        logging.error(msg)
        if not msg.endswith("."):
            msg += "."
        QtWidgets.QMessageBox.critical(self, self.tr("Error"), msg)

    def new_project(self):
        if not self._confirm_current_project_can_be_closed():
            return
        self._on_project_closed()

        dlg = DialogScenarioProperties(
            models.GeneralProperties.make_default(), self._working_dir, self
        )
        dlg.setWindowTitle(self.tr("New scenario"))
        # ask user to choose which schema to use for that new scenario
        dlg.enable_schema_selection()

        if dlg.exec_() != 0:
            schema_path = dlg.get_selected_schema_full_path()

            schema = models.Schema.load_yaml_file(schema_path)
            scenario = models.Scenario(schema, dlg.make_properties())
            self._controller.reset(scenario)

    def save_project(self) -> bool:
        if not self._controller.is_open:
            return False
        return self._do_save_project_as(self._controller.project_properties.file_path)

    def save_project_as(self) -> bool:
        if not self._controller.is_open:
            return False
        return self._do_save_project_as("")

    def _do_save_project_as(self, file_path: str) -> bool:
        assert self._controller.is_open

        if file_path == "":
            file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
                self,
                self.tr("Save scenario file"),
                self._working_dir.scenarios_dir,
                "Scenario file (*.yaml *.yml)",
            )
            if file_path == "":
                return False

        schema_is_valid = self._controller.revalidate_scenario()

        if not schema_is_valid:
            gen_quick_warning_modal(
                self.parentWidget(),
                "Schema Validation",
                "The scenario is valid according to the schema",
            )
            return False

        self._controller.save_to_file(file_path)
        self._graph_view.setSceneRect(self._controller.compute_scene_rect())
        self._controller.set_unsaved_changes(False)

        return True

    def close_project(self) -> None:
        if self._confirm_current_project_can_be_closed():
            self._on_project_closed()

    def show_open_scenario_file_dlg(self):
        if not self._confirm_current_project_can_be_closed():
            return
        logging.debug("show_open_scenario_file_dlg")

        file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            self.tr("Open scenario file"),
            self._working_dir.scenarios_dir,
            self.tr("Scenario (*.yaml *.yml)"),
        )
        if file_path != "":
            self.load_scenario_file(file_path)

    def _on_scenario_status_changed(self):
        logging.debug(
            "scenario status changed: agents={}, contracts={}".format(
                self._controller.agent_count, self._controller.contract_count
            )
        )

        is_open = self._controller.is_open
        self.ui.treeProjectStructure.setEnabled(is_open)
        self.ui.treeAttributes.setEnabled(is_open)
        self.ui.lineFilterPattern.setEnabled(is_open)
        self.fame_graphicsView.setEnabled(is_open)
        self.sliderZoomFactor.setEnabled(is_open)

        props = self._controller.project_properties
        if is_open:
            if props.file_path != "":
                self.ui.labelProjectName.setText(props.file_path)
            else:
                self.ui.labelProjectName.setText(self.tr("Unsaved scenario"))
        else:
            self.ui.labelProjectName.setText("")

        self.ui.actionSaveProject.setEnabled(props.has_unsaved_changes)

        self.ui.actionGeneralProperties.setEnabled(is_open)
        self.ui.actionConfigMenu.setEnabled(is_open)
        self.ui.actionMakeRunConfig.setEnabled(is_open)

        # update status bar
        if self._controller.agent_count > 0:
            if props.is_validation_successful:
                self._status_label_icon.setPixmap(QPixmap(":/icons/success-16px.png"))
                self._status_label_icon.setToolTip(
                    self.tr("Schema validation succeeded")
                )
            else:
                self._status_label_icon.setPixmap(QPixmap(":/icons/warning-16px.png"))
                all_errors = "\n".join(props.validation_errors)
                self._status_label_icon.setToolTip(
                    self.tr("Schema validation failed:\n{}".format(all_errors))
                )
        else:
            self._status_label_icon.clear()

    def load_scenario_file(self, file_path):
        self._on_project_closed()
        file_path = os.path.abspath(file_path)

        try:
            logging.info("opening scenario file {}".format(file_path))
            # TODO Future: prevent freezing UI at this place with Multithreading/Async IO -> Freezing place number 1

            scenario_model = models.ScenarioLoader.load_yaml_file(
                file_path, self._path_resolver,
                self._working_dir.root_dir)

            self._controller.init_scenario_model(scenario_model, file_path)
        except Exception as e:
            self._on_project_closed()
            self.display_error_msg("Failed to open scenario file: {}".format(e))
            return

        props = self._controller.project_properties
        if not props.is_validation_successful:
            QtWidgets.QMessageBox.warning(
                self,
                self.tr("Validation failure"),
                self.tr("The loaded scenario does not fulfill the schema:\n\n")
                + "\n".join(props.validation_errors),
            )
        self.ui.actionCloseProject.setVisible(True)

    def _on_edit_scenario_properties(self):
        dlg = DialogScenarioProperties(
            self._controller.scenario.general_properties, self._working_dir, self
        )

        dlg.setWindowTitle(self.tr("Scenario properties"))

        if dlg.exec_() != 0:
            edited_properties = dlg.make_properties().to_dict()
            original_properties = self._controller.scenario.general_properties.to_dict()

            if hash_dict(edited_properties) != hash_dict(original_properties):
                self._controller.save_project_properties_manipulation(
                    original_properties, edited_properties
                )

                self._controller.update_scenario_properties(dlg.make_properties())

    def _on_edit_fame_gui_settings(self):
        """Open up FAME GUI PRESETS and Preferences Settings Dialog"""
        dialog = EditFameDbValuesDialog(self._controller.scenario,

                                        self.handle_agent_color_change)
        dialog.exec_()

    def make_run_config(self):

        if self._controller.has_unsaved_changes:
            logging.info("Project unsaved, saving project to prepare for protobuf compilation")
            self._do_save_project_as(self._controller.project_properties.file_path)

        assert self._controller.can_export_protobuf
        scenario_name = os.path.basename(
            self._controller.project_properties.file_path
        ).replace(".yaml", "")
        output_path = "{}/{}.pb".format(self._working_dir.protobuf_dir, scenario_name)
        output_path = self._working_dir.make_relative_path(output_path)

        dlg = DialogScenarioProperties(
            self._controller.scenario.general_properties, self._working_dir, self
        )

        dlg.setWindowTitle(self.tr("Make run config"))
        dlg.enable_outputfile_selection(output_path)
        if dlg.exec() != 0:
            self._controller.update_scenario_properties(dlg.make_properties())
            self.save_project()
            output_path = self._working_dir.make_full_path(dlg.get_output_file_path())

            # display progress dialog
            progress_dlg = QtWidgets.QProgressDialog(self)
            progress_dlg.setLabelText(self.tr("Generating protobuf file..."))
            progress_dlg.setRange(0, 0)
            progress_dlg.setCancelButton(None)
            progress_dlg.show()
            QApplication.processEvents()

            try:
                models.write_protobuf_output(
                    self._controller.scenario, output_path, self._path_resolver
                )
                progress_dlg.close()
                QtWidgets.QMessageBox.information(
                    self,
                    self.tr("Success"),
                    self.tr(
                        "The following file was successfully generated:\n\n{}"
                    ).format(output_path),
                )

                self._controller.set_unsaved_changes(False)
            except Exception as e:
                progress_dlg.close()
                logging.error("failed to generate protobuf output: {}".format(e))
                QtWidgets.QMessageBox.critical(
                    self,
                    self.tr("Error"),
                    self.tr("Failed to generate the protobuf output.\n\n{}").format(e),
                )
            finally:
                progress_dlg.close()

    # prevent data loss when closing the main window
    def closeEvent(self, event):
        if not self._confirm_current_project_can_be_closed():
            event.ignore()
        else:
            event.accept()



import os

from famegui.models import Agent, Contract, GeneralProperties, Scenario, Schema

_dir = os.path.dirname(__file__)
_schema_path = os.path.abspath(
    os.path.join(_dir, "../../../testing_resources/schemas/schema-example.yaml")
)


def test_scenario_creation_from_scratch():
    # create empty scenario
    schema = Schema.load_yaml_file(_schema_path)
    props = GeneralProperties.make_default()
    s = Scenario(schema, props)
    assert len(s.agents) == 0
    assert len(s.contracts) == 0

    s.add_agent(Agent(1, "type1"))
    s.add_agent(Agent(2, "type2"))
    s.add_agent(Agent(3, "type3"))

    # TODO: fameio should reject duplicated agent
    # with pytest.raises(ValueError):
    #    s.add_agent(Agent(1, "type1"))
    # assert len(s.agents) == 3
    # assert len(s.contracts) == 0
    # assert s.agents[1].inputs == []
    # assert s.agents[1].outputs == []
    # assert s.agents[2].inputs == []
    # assert s.agents[2].outputs == []

    s.add_contract(Contract(1, 2, "contract 1->2", 10, 20))
    assert len(s.agents) == 3
    assert s.agents[1].inputs == []
    # TODO: enable these tests once Scenario.add_contract() has been updated to link inputs/outputs
    # assert s.agents[1].outputs == [2]
    # assert s.agents[2].inputs == [1]
    # assert s.agents[2].outputs == []
    assert len(s.contracts) == 1
    assert s.contracts[0].sender_id == 1
    assert s.contracts[0]._receiver_id == 2
    assert s.contracts[0].product_name == "contract 1->2"
    assert s.contracts[0].delivery_interval == 10
    assert s.contracts[0].first_delivery_time == 20
    assert s.contracts[0].expiration_time is None

    # accept contract with same sender / receiver
    s.add_contract(Contract(1, 1, "contract 1->1", 11, 21))
    assert len(s.contracts) == 2
    assert s.contracts[1].sender_id == 1
    assert s.contracts[1]._receiver_id == 1
    assert s.contracts[1].product_name == "contract 1->1"

    # bad contracts
    # TODO: fameio should reject those bad contracts
    # with pytest.raises(ValueError):
    #    s.add_contract(Contract(0, 1, "bad contract", 11, 21))  # bad sender
    # with pytest.raises(ValueError):
    #    s.add_contract(Contract(1, 0, "bad contract", 11, 21))  # bad receiver
    # assert len(s.contracts) == 2



import copy

from PySide6 import QtGui
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QFileDialog
from fameio.time import ConversionError

from famegui import models
from famegui.appworkingdir import AppWorkingDir
from famegui.generated.ui_dialog_scenario_properties import Ui_DialogScenarioProperties
from famegui.time_utils import (
    convert_gui_datetime_to_fame_time,
    convert_fame_time_to_gui_datetime,
)


class DialogScenarioProperties(QDialog):

    def __init__(
            self, props: models.GeneralProperties, workdir: AppWorkingDir, parent=None
    ):
        QDialog.__init__(self, parent)
        self._ui = Ui_DialogScenarioProperties()
        self._ui.setupUi(self)
        self._workdir = workdir
        self.hide_outputfile_selection()
        self.hide_schema_selection()

        # make sure to always have the required fields initialized
        self._props = copy.deepcopy(props)
        #        self._props.init_missing_values()

        # default button
        self._ui.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setDefault(True)

        # simulation params
        self._ui.lineEditStartTime.setDateTime(
            convert_fame_time_to_gui_datetime(self._props.simulation_start_time)
        )

        self._ui.lineEditStartTime.setDisplayFormat("dd/MM/yyyy hh:mm:ss")
        self._ui.lineEditStopTime.setDisplayFormat("dd/MM/yyyy hh:mm:ss")

        self._ui.lineEditStartTime.setCalendarPopup(True)
        self._ui.lineEditStopTime.setCalendarPopup(True)

        self._ui.lineEditStopTime.setDateTime(
            convert_fame_time_to_gui_datetime(self._props.simulation_stop_time)
        )
        self._ui.lineEditRandomSeed.setValidator(QtGui.QIntValidator())  # int64
        self._ui.lineEditRandomSeed.setText(str(self._props.simulation_random_seed))

        # output params

        # protobuf output file
        self._ui.buttonOutputPath.clicked.connect(self._on_select_output_path)

        # button status update
        self._connect_slots()
        self._update_ok_button_status()

    def _connect_slots(self):
        self._ui.lineEditStartTime.dateTimeChanged.connect(
            self._update_ok_button_status
        )
        self._ui.lineEditStopTime.dateTimeChanged.connect(self._update_ok_button_status)
        self._ui.fame_start_time.textChanged.connect(self._update_date_time_text_areas)
        self._ui.fame_stop_time.textChanged.connect(self._update_date_time_text_areas)

        self._ui.lineEditRandomSeed.textChanged.connect(self._update_ok_button_status)
        self._ui.lineEditOutputPath.textChanged.connect(self._update_ok_button_status)

    def hide_outputfile_selection(self):
        self._ui.groupBoxOutputFile.setEnabled(False)
        self._ui.groupBoxOutputFile.setVisible(False)
        self.adjustSize()

    def enable_outputfile_selection(self, default_path: str):
        self._ui.groupBoxOutputFile.setEnabled(True)
        self._ui.groupBoxOutputFile.setVisible(True)
        self._ui.lineEditOutputPath.setText(default_path)
        self.adjustSize()

    def hide_schema_selection(self):
        self._ui.groupBoxSchema.setEnabled(False)
        self._ui.groupBoxSchema.setVisible(False)
        self.adjustSize()

    def enable_schema_selection(self):
        self._ui.groupBoxSchema.setEnabled(True)
        self._ui.groupBoxSchema.setVisible(True)
        self._ui.comboBoxSchema.clear()
        self._ui.comboBoxSchema.addItems(self._workdir.list_existing_schema_files())
        self.adjustSize()

    def _on_select_output_path(self):
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Protobuf output file"),
            self._workdir.protobuf_dir,
            self.tr("Protobuf (*.pb)"),
        )
        if file_path != "":
            self._ui.lineEditOutputPath.setText(file_path)

    def accept(self):



        super().accept()

    def make_properties(self) -> models.GeneralProperties:
        # simulation
        datetime_string_start_time = self._ui.lineEditStartTime.text()
        datetime_string_stop_time = self._ui.lineEditStopTime.text()

        self._props.set_simulation_start_time(
            convert_gui_datetime_to_fame_time(datetime_string_start_time)
        )

        self._props.set_simulation_stop_time(
            convert_gui_datetime_to_fame_time(datetime_string_stop_time)
        )
        self._props.set_simulation_random_seed(int(self._ui.lineEditRandomSeed.text()))

        # output

        return self._props

    def get_output_file_path(self):
        assert self._ui.lineEditOutputPath.text() != ""
        return self._ui.lineEditOutputPath.text()

    def get_selected_schema_full_path(self) -> str:
        assert self._ui.comboBoxSchema.count() > 0
        selected_path = self._ui.comboBoxSchema.currentText()
        return self._workdir.make_full_path(selected_path)

    def _check_all_fields_are_valid(self):
        if (
                self._ui.lineEditStartTime.text() == ""
                or self._ui.lineEditStopTime.text() == ""
                or self._ui.lineEditRandomSeed.text() == ""
        ):
            return False
        if (
                self._ui.lineEditOutputPath.isVisible()
                and self._ui.lineEditOutputPath.text() == ""
        ):
            return False
        return True

    def _update_date_time_text_areas(self):
        fame_stop_time: str = self._ui.fame_stop_time.text()
        if fame_stop_time == "":
            return

        gui_stop_time = convert_fame_time_to_gui_datetime(int(fame_stop_time))
        self._ui.lineEditStopTime.setDateTime(gui_stop_time)
        fame_start_time: str = self._ui.fame_start_time.text()
        if fame_start_time == "":
            return
        gui_start_time = convert_fame_time_to_gui_datetime(int(fame_start_time))
        self._ui.lineEditStartTime.setDateTime(gui_start_time)

    def _update_fame_time_text_areas(self):

        self._ui.fame_stop_time.setStyleSheet("QLabel {color: black}")
        self._ui.fame_start_time.setStyleSheet("QLabel {color: black}")

        try:
            fame_start_time = convert_gui_datetime_to_fame_time(
                self._ui.lineEditStartTime.text()
            )
            self._ui.fame_start_time.setText(str(fame_start_time))

        except ConversionError as c_ex:

            self._ui.fame_start_time.setText(str(c_ex))
            self._ui.fame_start_time.setStyleSheet("QLabel {color: red}")

        try:
            fame_stop_time = convert_gui_datetime_to_fame_time(
                self._ui.lineEditStopTime.text()
            )
            self._ui.fame_stop_time.setText(str(fame_stop_time))

        except ConversionError as c_ex:
            self._ui.fame_stop_time.setText(str(c_ex))
            self._ui.fame_stop_time.setStyleSheet("QLabel {color: red}")

    def _update_ok_button_status(self):

        self._ui.lineEditStopTime.blockSignals(True)
        self._ui.lineEditStartTime.blockSignals(True)
        self._update_fame_time_text_areas()
        self._ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(
            self._check_all_fields_are_valid()
        )

        self._ui.lineEditStopTime.blockSignals(False)
        self._ui.lineEditStartTime.blockSignals(False)



from PySide6.QtGui import QPainter, QColor
from PySide6.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QFrame
from fameio.input.schema import AttributeSpecs
from icecream import ic

from famegui.ui.block_item_wrapper import BlockItemWrapper
from famegui.ui.block_type_wrapper import BlockTypeWrapper


class ListItem(QFrame):
    """Wrapper for all types of Sub Widgets of a List Widget."""

    def __init__(self, attr_name: str, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.attr_name = attr_name

        self.card_wrapper_layout = QVBoxLayout()
        self.setLayout(self.card_wrapper_layout)

        color = "rgba(255, 221, 148, 0.7)"
        self.setStyleSheet(
            f"ListItem {{background-color: {color};  border-radius: 5px;}}"
        )

        self.input_panel = None

    def get_attr_name(self):
        """Get the attribute short name."""
        return self.attr_name

    def set_panel_widget(self, widget):
        """Set the input panel widget."""
        self.input_panel = widget

        uniq_widget = f"myUniqueWidget-{id(self.input_panel)}"
        self.input_panel.setObjectName(uniq_widget)

        self.input_panel.update()
        self.update()

        self.layout().addWidget(widget)

    def get_attr(self) -> AttributeSpecs:
        """Get the AttributeSpecs of the input widget."""

        for i in range(self.layout().count()):
            item = self.layout().itemAt(i)
            if item:
                widget = item.widget()
                if isinstance(widget, BlockTypeWrapper):
                    widget: BlockTypeWrapper
                    return widget.attr_specs
                if isinstance(widget, BlockItemWrapper):
                    widget: BlockItemWrapper
                    return widget.get_attr_spec()
                if widget.__class__.__name__ == "InputPanelWrapper":
                    return widget.get_attribute_spec()

    def get_attribute_full_name(self) -> str:
        """Get the attribute full name."""
        for i in range(self.layout().count()):
            item = self.layout().itemAt(i)
            if item:
                widget = item.widget()
                if isinstance(widget, BlockTypeWrapper):
                    widget: BlockTypeWrapper
                    return widget.attr_specs.full_name
                if isinstance(widget, BlockItemWrapper):
                    widget: BlockItemWrapper
                    return widget.get_attr_full_name()
                if widget.__class__.__name__ == "InputPanelWrapper":
                    return widget.get_attribute_full_name()

    def _process_block_type_input(self, widget, data):
        inner_data = widget.get_data()
        for key, value in widget.get_data().items():
            inner_data[key] = value
        data[self.attr_name] = inner_data

    def _process_block_item_type_input(self, widget, data):
        for key, value in widget.get_data().items():
            data[key] = value

    def _process_input_panel_wrapper_input(self, widget, data):
        inner_data = widget.get_input_value()

        if isinstance(inner_data, dict):
            for key, value in inner_data.items():
                data[key] = value
        if isinstance(inner_data, list):
            spec = widget.get_attribute_spec()


            sepc: AttributeSpecs
            if spec.is_list:
                data[self.attr_name] = inner_data
                return
            for sub_data in inner_data:
                if isinstance(sub_data, dict):
                    for key, value in sub_data.items():
                        data[key] = value
                    continue
                data[self.attr_name] = inner_data

        return

    def get_data(self):
        """Group and gather all the data from the input widget."""

        data = {}

        for i in range(self.layout().count()):
            item = self.layout().itemAt(i)
            if item:
                widget = item.widget()

                if widget:
                    if isinstance(widget, QLineEdit):
                        data[self.attr_name] = widget.text()

                        continue

                    if isinstance(widget, BlockTypeWrapper):
                        self._process_block_type_input(widget, data)

                        continue

                    if isinstance(widget, BlockItemWrapper):
                        self._process_block_item_type_input(widget, data)

                        continue
                    if widget.__class__.__name__ == "InputPanelWrapper":
                        self._process_input_panel_wrapper_input(widget, data)

        return data



from famegui.database.db_loader.db_manager import get_single_row, check_entry_exists
from famegui.database.settings_db.init_db import init_db_session


def check_if_agent_type_has_color(agent_type_name: str):
    exist = False
    with init_db_session() as session:
        exist = check_entry_exists(session, "agent_colors_second_table", "agent_type_name", agent_type_name)

    return exist







UPDATED_CONTRACT_KEY = "updated_contract"
OLD_CONTRACT_KEY = "old_contract"

CONTRACT_SENDER_ID = "senderid"
CONTRACT_RECEIVER_ID = "receiverid"
CONTRACT_PRODUCT_NAME = "productname"
CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS = "deliveryintervalinsteps"


from PySide6.QtWidgets import QLabel


class QFameBoldLabel(QLabel):
    """A QLabel with bold font"""

    def __init__(self, text: str, font_size=10, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setText(text)
        font = self.font()
        font.setPointSize(font_size)
        font.setBold(True)
        self.setFont(font)


class DescriptionLabel(QLabel):
    """A QLabel with italic font for descriptions"""

    def __init__(self, text: str, font_size=10, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setText(text)
        font = self.font()
        font.setPointSize(font_size)
        font.setItalic(True)
        self.setFont(font)



import logging
import os

from PySide6 import QtWidgets, QtCore, QtUiTools
from PySide6.QtCore import QFile
from PySide6.QtWidgets import QLabel, QLineEdit, QHBoxLayout, QPushButton, QComboBox, QDateTimeEdit, QVBoxLayout, \
    QScrollArea, QWidget

from famegui.config.config import CUSTOM_DELIVERY_STEPS_TIME_OPTION
from famegui.config.contract_consts import PRESET_DURATION_IN_SECONDS
from famegui.config.runtime_consts import UPDATED_CONTRACT_KEY, OLD_CONTRACT_KEY, \
    CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS, CONTRACT_RECEIVER_ID, CONTRACT_SENDER_ID, CONTRACT_PRODUCT_NAME
from famegui.database.prebuild_queries.interval_presets import get_delivery_interval_presets
from famegui.maincontroller import MainController
from famegui.models import Contract, Scenario, Agent
from famegui.ui.fame_input_panels import (
    get_string_chooser_panel,
    FameCalendarInputPanelWidget,
)
from famegui.ui.fame_ui_elements import QFameBoldLabel
from famegui.ui.helper import gen_preset_interval_chooser
from famegui.ui.quick_modals import gen_quick_warning_modal
from famegui.utils import contract_fields_to_natural_types

DISPLAY_DATE_FORMAT = "yyyy-MM-dd"

DISPLAY_DATETIME_FORMAT = "yyyy-MM-dd HH:mm:ss"


class MultiEditContractDialog(QtWidgets.QDialog):
    """Dialog for editing a contract"""

    on_contract_edited = QtCore.Signal(dict)

    def __init__(
            self,
            selected_contracts: [Contract],
            scenario: Scenario,
            controller: MainController,
            on_contract_edited_slot,
            parent=None,
    ):
        super(MultiEditContractDialog, self).__init__(parent)

        # Set up the user interface from Designer
        self._selected_contracts = selected_contracts
        self._scenario = scenario
        self._controller = controller
        self.on_contract_edited.connect(on_contract_edited_slot)
        self._product_changed = False

        self._delivery_interval_presets = get_delivery_interval_presets()

        # Load UI
        loader = QtUiTools.QUiLoader()
        file = QFile(os.path.join(os.path.dirname(__file__), "edit_contract_dialog.ui"))
        file.open(QtCore.QFile.OpenModeFlag.ReadOnly)
        self.ui = loader.load(file, self)

        file.close()
        self._input_fields = []
        self.ui.verticalLayout: QtWidgets.QVBoxLayout

        self._init_static_ui()
        self._generate_input_items()
        self.create_btn_panel()

        # Create a scroll area
        selected_contracts_panel_scroll_area = QScrollArea(self)
        selected_contracts_panel_scroll_area.setWidgetResizable(True)
        selected_contracts_panel_scroll_area.setMaximumHeight(90)  # Cap the height
        selected_contracts_panel_scroll_area.setMinimumHeight(90)  # Fix the height

        # Create the child widget for the scroll area
        selected_contracts_panel = QWidget()
        selected_contracts_panel.setStyleSheet("background-color:#E0E0E0")

        # Set a vertical layout for the child widget
        selected_contracts_panel_layout = QVBoxLayout(selected_contracts_panel)

        # Add contract labels to the layout
        for selected_contract in self._selected_contracts:
            label = QLabel(
                f"Sender Id: {selected_contract.sender_id} - Receiver Id: {selected_contract.receiver_id} ({selected_contract.product_name})"
            )
            selected_contracts_panel_layout.addWidget(label)

        # Set the child widget as the content of the scroll area
        selected_contracts_panel_scroll_area.setWidget(selected_contracts_panel)

        # Add the scroll area to the main layout
        self.ui.verticalLayout.insertWidget(0, selected_contracts_panel_scroll_area)

        # Add a note below the scroll area
        self.ui.verticalLayout.insertWidget(1, QLabel(
            "! Only fill the fields you wish to apply on all selected contracts"
        ))

        self.ui.exec_()

    def setup_layout(self):
        """Set up the layout frame of the dialog"""
        self.ui.verticalLayout.setContentsMargins(20, 20, 20, 20)
        self.ui.verticalLayout.setSpacing(20)

    def _set_product_changed(self):
        """Helper to flag that the product was changed """
        self._product_changed = True

    def _init_static_ui(self):
        """Set up the static UI elements of the dialog"""
        self.setup_layout()

    def _construct_product_chooser(self, layout, field_name):
        """construct product chooser widget"""

        sender_type = self._scenario.schema.agent_type_from_name(
            self._controller.get_agent_ctrl(self._selected_contracts[0].sender_id).type_name
        )

        sender_type_products = sender_type.products

        inner_layout, comboBox = get_string_chooser_panel(sender_type_products, Contract.KEY_PRODUCT,
                                                          preset_value="",
                                                          on_input_changed=lambda text: self._set_product_changed())

        layout.addLayout(inner_layout)
        self._input_fields.append((field_name, comboBox))

    def _construct_date_time_chooser(self, layout, field_name, contract_data):
        """construct product time chooser widget"""

        saved_fame_datetime = None

        if contract_data is not None:
            if field_name in contract_data:
                saved_fame_datetime = contract_data[field_name]

        fame_calender_input_panel = FameCalendarInputPanelWidget(
            self, field_name, saved_fame_datetime
        )
        (
            inner_layout,
            date_time_edit,
        ) = fame_calender_input_panel.get_input_related_widgets()
        layout.addWidget(fame_calender_input_panel)

        self._input_fields.append((field_name, date_time_edit))

    def on_preset_changed(self, index, selected_option: str):
        """
        Handles the change event when a new preset is selected
        """
        # Do something with the selected preset

        for field_name, input in self._input_fields:

            if field_name == CONTRACT_FIELD_DELIVERY_INTERVAL_IN_STEPS:
                if selected_option == "Custom":
                    continue
                input.setText(self._delivery_interval_presets[index][PRESET_DURATION_IN_SECONDS])


    def _construct_default_input_field(self, layout, field_name, contract_data):
        """construct default input field"""
        label = QFameBoldLabel(text=field_name)
        label.setMinimumWidth(int(self.width() * 0.3))  # Set the maximum width to 30% of the dialog's width
        label.setMaximumWidth(int(self.width() * 0.3))  # Set the maximum width to 30% of the dialog's width
        label.setWordWrap(True)  # Enable word wrap

        input_field = QLineEdit()

        if contract_data is not None:
            if field_name in contract_data:
                input_field.setText(str(contract_data[field_name]))

        inner_layout = QHBoxLayout()
        inner_layout.addWidget(label)

        choices = [item["preset_label"] for item in self._delivery_interval_presets]

        choices_bundles = [item for item in self._delivery_interval_presets]

        interval_preset_chooser = gen_preset_interval_chooser(choices, self.on_preset_changed,
                                                              input_field, choices_bundles,
                                                              CUSTOM_DELIVERY_STEPS_TIME_OPTION)
        if field_name.lower() == "deliveryintervalinsteps":
            inner_layout.addWidget(interval_preset_chooser)

        inner_layout.addWidget(input_field)

        layout.addLayout(inner_layout)
        self._input_fields.append((field_name, input_field))

    def _generate_input_items(self):
        """Entry point for generating input items"""

        layout = self.ui.verticalLayout

        # Create input rows dynamically based on the item list
        for field_name in self._selected_contracts[0].get_fields_to_edit():
            if field_name == Contract.KEY_PRODUCT:
                self._construct_product_chooser(layout, field_name)

                continue

            field_type = self._selected_contracts[0].get_field_type(field_name)

            if field_type == self._selected_contracts[0].FIELD_TYPE_DATE_TIME:
                self._construct_date_time_chooser(layout, field_name, None)

                continue
            self._construct_default_input_field(layout, field_name, None)

    def create_btn_panel(self):
        """Construct the button panel"""
        cancel_button = QPushButton("Cancel", self)
        confirm_button = QPushButton("Confirm", self)
        cancel_button.clicked.connect(self.cancel)
        confirm_button.clicked.connect(self.confirm)

        # Create a QHBoxLayout for the buttons
        button_layout = QHBoxLayout()

        # Add the buttons to the button layout
        button_layout.addWidget(cancel_button)
        button_layout.addWidget(confirm_button)

        self.ui.verticalLayout.addLayout(button_layout)

    def init_ui(self):
        input = QLineEdit(self)
        self.ui.verticalLayout.addWidget(input)

    def cancel(self):
        """Cancel the applied changes and close the dialog"""
        self.ui.close()

    def _check_if_agent_exists(self, agent_id) -> bool:

        if not self._scenario.agent_exists(agent_id):
            gen_quick_warning_modal(
                self, "Agent does not exist", "Agent with id " + str(agent_id) + " does not exist"
            )
            return False

        return True

    def _check_for_product_support(self, agent: Agent) -> bool:

        if not self._scenario.schema.agent_supports_product(
                self._scenario.schema.agent_types[agent.type_name], self._selected_contracts[0].product_name
        ):
            message = f"Agent with id '{str(agent.id)}' does not support the product: {self._selected_contracts[0].product_name}"
            gen_quick_warning_modal(self, "Sender does not support product", message)
            return False
        return True

    def confirm(self):
        """Confirm the applied changes and close the dialog"""
        data_dict = {}

        for field_name, input_field in self._input_fields:
            if isinstance(input_field, QLineEdit):
                input_field: QLineEdit
                data_dict[field_name] = input_field.text()

                try:
                    input_data = int(input_field.text())
                except Exception as exce:
                    logging.debug(F"OPTIONAL FIELD NOT SET {str(exce)}")
                    continue

                if field_name == Contract.KEY_SENDER or field_name == Contract.KEY_SENDER:
                    if not self._scenario.agent_exists(input_data):
                        if not self._check_if_agent_exists(input_data):
                            return

                    agent: Agent = self._scenario.get_agent_by_id(input_data)

                    if not self._check_for_product_support(agent):
                        return

            if isinstance(input_field, QDateTimeEdit):
                input_field: QDateTimeEdit
                data_dict[field_name] = input_field.dateTime().toString(DISPLAY_DATETIME_FORMAT)
            if isinstance(input_field, QComboBox):
                input_field: QComboBox
                data_dict[field_name] = input_field.currentText()

        for contract in self._selected_contracts:
            old_agent_dict = contract.to_dict()

            dict_to_send = {
                OLD_CONTRACT_KEY: old_agent_dict.copy(),
            }

            for key, value in data_dict.items():

                if value is None:
                    continue
                if str(value) == "":
                    continue
                if key == CONTRACT_PRODUCT_NAME:
                    if not self._product_changed:
                        continue

                if key == CONTRACT_RECEIVER_ID:
                    continue
                if key == CONTRACT_SENDER_ID:
                    continue
                old_agent_dict[key] = value

            old_agent_dict = contract_fields_to_natural_types(
                old_agent_dict, contract
            )
            updated_contract = Contract.from_dict(old_agent_dict)

            dict_to_send[UPDATED_CONTRACT_KEY] = updated_contract.to_dict()

            self._scenario.update_contract(
                old_contract=contract, updated_contract=updated_contract
            )

            self.on_contract_edited.emit(dict_to_send)

        self._controller.set_unsaved_changes(True)

        self.ui.accept()

    @staticmethod
    def select_date(date, line_edit):
        selected_date = date.toString(DISPLAY_DATE_FORMAT)
        line_edit.setText(selected_date)



import typing
from enum import Enum

from famegui.database.db_loader.db_manager import get_single_row
from famegui.database.settings_db.init_db import init_db_session

DB_SETTING_TABLE_NAME = "settings"
DB_SETTING_NAME_FIELD = "setting_name"


class FameContractGraphItemSettings(object):
    CONTRACT_LINE_ANGLE = "contract_line_angle"
    CONTRACT_LINE_WIDTH = "contract_line_width"


class FameGeneralRuntimeSetting:
    AMOUNT_OF_UNDO_STEPS = "amount_of_undo_steps"


def get_setting_value(setting_name_entry: str, fallback_value: typing.Union[str, int]) -> str:
    """
    Retrieve the value of a setting by its name from the settings table.
    If no setting is found, the fallback value is returned.

    Args:
        setting_name_entry (str): The name of the setting to search for.
        fallback_value (str): The value to return if the setting is not found.

    Returns:
        str: The value of the setting if found, otherwise the fallback value.
    """
    with init_db_session() as session:
        result = get_single_row(session, DB_SETTING_TABLE_NAME, DB_SETTING_NAME_FIELD, setting_name_entry, ["value"])

    # Return the result if found, otherwise return the fallback value
    if result and "value" in result:
        return result["value"]


    else:
        return fallback_value


if __name__ == "__main__":

    value = int(get_setting_value(FameContractGraphItemSettings.CONTRACT_LINE_ANGLE,
                                  20))



import logging
import os
import typing

from fameio.input.resolver import PathResolver

from famegui.appworkingdir import AppWorkingDir


class FameGuiPathResolver(PathResolver):
    def __init__(self, work_dir: AppWorkingDir):
        self._work_dir = work_dir

    def resolve_yaml_imported_file_pattern(
        self, root_path: str, file_pattern: str
    ) -> typing.List[str]:
        logging.debug(
            "resolving yaml imported file pattern '{}' from root path {}".format(
                file_pattern, root_path
            )
        )
        # first, try to locate the file(s) via the default behaviour
        result = super().resolve_yaml_imported_file_pattern(root_path, file_pattern)
        if len(result) > 0:
            logging.info("resolved yaml imported file(s) to {}".format(result))
            return result

        logging.warning(
            "failed to locate yaml imported file(s) from pattern {}".format(
                file_pattern
            )
        )
        return []

    def resolve_series_file_path(self, file_name: str) -> typing.Optional[str]:
        if os.path.isabs(file_name):
            return file_name

        # try first with the default resolver
        file_path = super().resolve_series_file_path(file_name)
        if file_path:
            return file_path

        # try again in the working dir
        file_path = self._work_dir.find_existing_child_file(file_name)
        if file_path:
            return file_path

        logging.warning("failed to locate timeseries file '{}'".format(file_name))
        return file_path



import hashlib
import ujson as json


def hash_dict(d):
    """Generate a consistent hash for a dictionary."""
    # Serialize the dictionary into a sorted JSON string
    serialized = json.dumps(d, sort_keys=True)
    # Hash the serialized string
    return hashlib.sha256(serialized.encode()).hexdigest()



from enum import Enum

from fameio.input.schema import AttributeType, AttributeSpecs


class StringAttributeType(Enum):
    BLOCK = "AttributeType.BLOCK"
    TIME_SERIES = "AttributeType.TIME_SERIES"
    ENUM = "AttributeType.ENUM"
    INTEGER = "AttributeType.INTEGER"
    LONG = "AttributeType.LONG"
    DOUBLE = "AttributeType.DOUBLE"
    STRING = "AttributeType.STRING"
    STRING_SET = "AttributeType.STRING_SET"


def get_full_attr_names(data, depth=None, current_depth=0):
    """
    Recursively extract full attribute names from nested data structure.

    Args:
        data: The nested data structure to process
        depth: Maximum depth to traverse (None for unlimited)
        current_depth: Current recursion depth

    Returns:
        list: Flat list of full attribute names
    """
    result = []

    # Return if we've reached max depth
    if depth is not None and current_depth >= depth:
        return result

    # Handle dictionary case
    if isinstance(data, dict):
        # Check if it's an attribute with 'attr' key
        if 'attr' in data:
            # Add the attribute name if it exists
            if isinstance(data['attr'], str):
                result.append(data['attr'])

            # Recursively process nested attributes
            if data.get('has_nested_attributes'):
                for key, value in data.get('attr', {}).items():
                    if isinstance(value, dict):
                        result.extend(get_full_attr_names(value, depth, current_depth + 1))

        # Process all dictionary values
        for value in data.values():
            if isinstance(value, (dict, list)):
                result.extend(get_full_attr_names(value, depth, current_depth + 1))

    # Handle list case
    elif isinstance(data, list):
        for item in data:
            result.extend(get_full_attr_names(item, depth, current_depth + 1))

    return result



def flatten_dict(full_dict: dict, d: dict, parent_key: str = '', sep: str = '.') -> dict:
    """
    Flatten a nested dictionary using dot notation, handling dictionaries and lists.

    Args:
        full_dict (dict): The original full dictionary for reference.
        d (dict or list): The current level dictionary or list to flatten.
        parent_key (str): The base key string accumulated so far.
        sep (str): The separator to use between keys.

    Returns:
        dict: A flattened dictionary with dot-separated keys.
    """
    items = {}
    end_node_keywords = ["mandatory", "list"]

    if isinstance(d, dict):
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k

            # Check for end node keywords to potentially skip or handle specially
            if any(keyword in k.lower() for keyword in end_node_keywords):
                # Depending on your specific needs, you can choose to skip or process differently
                pass  # Currently, we're processing all keys

            # Recursively process dictionaries and lists
            if isinstance(v, dict):
                items.update(flatten_dict(full_dict, v, new_key, sep))
            elif isinstance(v, list):
                items.update(flatten_dict(full_dict, v, new_key, sep))
            else:
                items[new_key] = v

    elif isinstance(d, list):
        new_key = parent_key

        items[new_key] = [sub_item for sub_item in d]

    return items



def get_nested_value(data, keys_list):
    for key in keys_list:
        if not isinstance(data, dict) and not isinstance(data, list):
            return data

        if isinstance(data, list):
            data = [sub_data[key] if key in sub_data else sub_data for sub_data in data]

            continue

        data = data[key] if key in data else data

    return data

def get_sub_attr(attr, flat_values, type_name):
    sub_dict = {}
    none_if_dict_or_list = lambda x: None if isinstance(x, dict) else x
    for key, item in attr:

        if not item.attr_type == AttributeType.BLOCK:
            item: AttributeSpecs

            value_key = item.full_name.replace(type_name + ".", "")
            alt_value = None

            if value_key.__contains__("."):
                keys = value_key.split(".")
                alt_value = get_nested_value(flat_values, keys)

                alt_value = alt_value if not isinstance(alt_value, dict) else None

            get_value_from_value_dict = lambda value_item_key: none_if_dict_or_list(
                flat_values[value_item_key]) if value_item_key in flat_values else None

            new_item = {
                "type": str(item.attr_type),
                "attr": item.full_name,
                "is_list": item.is_list,
                "flat_values": flat_values,
                "has_nested_attributes": item.has_nested_attributes,
                "is_mandatory": item.is_mandatory,
                "value": get_value_from_value_dict(value_key) if flat_values is not None else None,
                "value_key": value_key
            }

            if item.attr_type == AttributeType.ENUM:
                new_item["options"] = item.values

            if alt_value:
                new_item["value"] = none_if_dict_or_list(alt_value)

            sub_dict[key] = new_item

            continue
        value_key = item.full_name.replace(type_name + ".", "")

        alt_value = None

        get_value_from_value_dict = lambda value_item_key: none_if_dict_or_list(
            flat_values[value_item_key]) if value_item_key in flat_values else None

        new_item = {
            "type": str(item.attr_type),
            "attr_name": item.full_name,
            "attr": get_sub_attr(item.nested_attributes.items(), flat_values, type_name),
            "is_list": item.is_list,
            "alt_value": alt_value,
            "list_values": None,
            "is_mandatory": item.is_mandatory,
            "has_nested_attributes": item.has_nested_attributes,
            "flat_values": flat_values,
            "value": get_value_from_value_dict(value_key) if flat_values is not None else None,

        }
        if alt_value:
            new_item["value"] = alt_value if not isinstance(alt_value, dict) else None

        sub_dict[key] = new_item

    return sub_dict


def get_string_between_end_and_last_dot(input_string):
    if '.' in input_string:
        return input_string[input_string.rfind('.') + 1:]
    return


def build_simple_attr_dict(attr_spec: AttributeSpecs,
                           flat_full_name_labeled_values: dict,
                           agent_type_name: str, is_list_item=False):
    data_out = {}

    if attr_spec.has_nested_attributes:
        for key, item in attr_spec.nested_attributes.items():
            item: AttributeSpecs

            if item.has_nested_attributes:
                res_out = build_simple_attr_dict(item, flat_full_name_labeled_values,
                                                 agent_type_name,
                                                 is_list_item=item.is_list)

                data_out[key] = res_out
                continue

            full_name_without_agent_name = item.full_name.replace(agent_type_name + ".", "")

            data_out[key] = flat_full_name_labeled_values[
                full_name_without_agent_name] if full_name_without_agent_name in flat_full_name_labeled_values else None

        return data_out

    full_name_without_agent_name = attr_spec.full_name.replace(agent_type_name + ".", "")

    data_out[full_name_without_agent_name] = flat_full_name_labeled_values[
        full_name_without_agent_name] \
        if full_name_without_agent_name in flat_full_name_labeled_values else None

    return data_out



import logging
import os
import pytest
import typing

import yaml

from famegui.models import ScenarioLoader
import fameio.input.loader as fameio

_dir = os.path.dirname(__file__)


class CustomPathResolver(fameio.PathResolver):
    def resolve_yaml_imported_file_pattern(
            self, root_path: str, file_pattern: str
    ) -> typing.List[str]:
        # adjust root path to where the schema is stored
        root_path = os.path.abspath(_dir + "../../../../testing_resources")
        result = super().resolve_yaml_imported_file_pattern(root_path, file_pattern)
        assert len(result) > 0
        return result


def test_load_valid_yaml_file():
    path_resolver = CustomPathResolver()
    fame_work_dir = os.path.abspath(_dir + "../../../../testing_resources")

    s = ScenarioLoader.load_yaml_file(_dir + "/files/scenario.yaml", path_resolver, fame_work_dir)

    assert len(s.agents) == 50
    assert s.agents[0].type_name == "DayAheadMarketSingleZone"
    assert s.agents[0].display_id == "#1"
    assert s.agents[1].type_name == "CarbonMarket"
    assert s.agents[1].display_id == "#3"
    assert s.agents[2].type_name == "FuelsMarket"
    assert s.agents[2].display_id == "#4"
    assert s.agents[3].type_name == "DemandTrader"
    assert s.agents[3].display_id == "#100"

    assert len(s.contracts) == 294
    assert s.contracts[0].sender_id == 2000
    assert s.contracts[0].display_receiver_id == "#500"
    assert s.contracts[0].product_name == "PowerPlantPortfolio"
    # TODO: enable these tests once Scenario.add_contract() has been updated to link inputs/outputs
    # assert s.agents[3].outputs == [3]
    # assert s.agents[2].inputs == [4]

    assert s.contracts[1].display_sender_id == "#2001"
    assert s.contracts[1].receiver_id == 501
    assert s.contracts[1].product_name == "PowerPlantPortfolio"
    # assert s.agents[5].outputs == [4]
    # assert s.agents[4].inputs == [5]

    assert s.contracts[2].sender_id == 2002
    assert s.contracts[2].receiver_id == 502
    assert s.contracts[2].product_name == "PowerPlantPortfolio"
    # assert s.agents[2].outputs == [1]
    # assert s.agents[1].inputs == [2]


def is_yaml_empty(file_path):
    try:
        with open(file_path, 'r') as file:
            content = file.read().strip()  # Remove any leading/trailing whitespace
            if not content:
                return True  # File is empty or only whitespace

            # Attempt to parse the YAML content
            yaml_data = yaml.safe_load(content)
            if yaml_data is None:  # YAML content is valid but empty
                return True
    except Exception as e:
        logging.info(f"Error reading YAML file: {e}")
    return False


def test_load_empty_yaml_file():
    # we accept empty scenario files (to not block the user in the GUI)
    path_resolver = CustomPathResolver()
    fame_work_dir = os.path.abspath(_dir + "../../../../testing_resources")

    empty_scenario_path = _dir + "/files/empty.yaml"

    assert is_yaml_empty(empty_scenario_path) == True

    # todo: talk to DLR:: empty scenario is not loading due to lower_keys function

    ScenarioLoader.load_yaml_file(_dir + "/files/empty.yaml", path_resolver, fame_work_dir)


def test_load_single_agent_yaml_file():
    path_resolver = CustomPathResolver()
    fame_work_dir = os.path.abspath(_dir + "../../../../testing_resources")

    s = ScenarioLoader.load_yaml_file(_dir + "/files/single_agent.yaml", path_resolver, fame_work_dir)
    assert len(s.agents) == 1
    assert len(s.contracts) == 0


def test_load_single_contract_yaml_file():
    path_resolver = CustomPathResolver()
    fame_work_dir = os.path.abspath(_dir + "../../../../testing_resources")
    # TODO: this scenario file should be rejected (contract references missing agents)
    s = ScenarioLoader.load_yaml_file(
        _dir + "/files/single_contract.yaml", path_resolver, fame_work_dir
    )
    assert len(s.agents) == 0
    assert len(s.contracts) == 1


def test_load_invalid_path():
    path_resolver = CustomPathResolver()
    fame_work_dir = os.path.abspath(_dir + "../../../../testing_resources")
    with pytest.raises(FileNotFoundError):
        ScenarioLoader.load_yaml_file(_dir + "/invalid_path", path_resolver, fame_work_dir)



import os

PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))

CACHE_DIR = os.path.join(PROJECT_DIR, "cache")

os.makedirs(CACHE_DIR, exist_ok=True)

CACHE_TRACKING_DB_DIR = os.path.join(CACHE_DIR, "db")

CACHE_TRACKING_DB_PATH = os.path.join(CACHE_TRACKING_DB_DIR, "db.json")



import os

import pytest
import ujson
from fameio.input.resolver import PathResolver
from famegui import models
from famegui.config.config import DEFAULT_FAME_WORK_DIR
from famegui.data_manager.string_set_manager import get_pre_defined_string_set_values_from_scenario

@pytest.fixture
def scenario_file_path():
    """Fixture to ensure the scenario file exists and provide its path."""
    assert os.path.exists(DEFAULT_FAME_WORK_DIR), f"Path does not exist: {DEFAULT_FAME_WORK_DIR}"
    file_path = os.path.join(DEFAULT_FAME_WORK_DIR, "scenario.yaml")
    assert os.path.exists(file_path), f"Scenario file does not exist: {file_path}"
    return file_path


def test_load_scenario_model(scenario_file_path):
    """Test to load the scenario model and verify its integrity."""
    scenario_model = models.ScenarioLoader.load_yaml_file(
        scenario_file_path,
        PathResolver(),
        DEFAULT_FAME_WORK_DIR
    )
    assert scenario_model is not None, "Failed to load scenario model"

def test_pre_defined_string_set_values(scenario_file_path):
    """Test to retrieve pre-defined string set values from the scenario model."""
    scenario_model = models.ScenarioLoader.load_yaml_file(
        scenario_file_path,
        PathResolver(),
        DEFAULT_FAME_WORK_DIR
    )
    string_set_data_dict = get_pre_defined_string_set_values_from_scenario(scenario_model)
    assert string_set_data_dict is not None, "Failed to retrieve string set data"
    assert isinstance(string_set_data_dict, dict), "String set data should be a dictionary"



import os.path
from typing import List

from PySide6 import QtWidgets, QtCore
from PySide6.QtCore import QFile, Qt, Signal
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QDialogButtonBox

from famegui.agent_controller import AgentController
from famegui.maincontroller import MainController
from famegui.models import Scenario


class AgentTreeItem(QtWidgets.QTreeWidgetItem):
    """Tree item for agent that shows agent metadata with a checkbox"""

    def __init__(
            self,
            parent: QtWidgets.QTreeWidget,
            agent_id: int,
            agent_type: str,
            amount_of_contracts: int,
            amount_of_connected_agents: int,
    ):
        self._agent_id = agent_id
        self._agent_type = agent_type
        QtWidgets.QTreeWidgetItem.__init__(self, parent)

        self.setFlags(self.flags() | Qt.ItemIsUserCheckable)
        self.setCheckState(0, Qt.Checked)

        self._display_error = lambda has_error: None

        self.setText(0, str(agent_id))
        self.setText(1, str(agent_type))
        self.setText(2, str(amount_of_contracts))
        self.setText(3, str(amount_of_connected_agents))


class DeletionDialog(QtCore.QObject):
    """Dialog for Agent deletion"""

    def iterate_items(self, parent_item, func):
        pass

    def _get_checked_agents(self, parent_item) -> List[int]:
        """Returns list of by user selected agents"""
        return [
            child_item.text(0)
            for child_item in list(
                filter(lambda x: x.checkState(0) == Qt.Checked, self._tree_items)
            )
        ]

    def __init__(
            self,
            agents_arr: [AgentController],
            scenario: Scenario,
            working_dir,
            parent_widget,
            main_controller,
    ):
        super().__init__()
        self._tree_items = []
        self.scenario: Scenario = scenario
        self._main_controller: MainController = main_controller
        self._agents: [AgentController] = agents_arr

        ui_file = QFile(
            os.path.join(os.path.dirname(__file__), "delete_agent_dialog.ui")
        )
        ui_file.open(QFile.ReadOnly)
        self._loader = QUiLoader()

        self._ui = self._loader.load(ui_file)
        self.init_ui()
        self._ui.exec_()

    def confirm_button_clicked(self):
        """Performs user chosen actions"""
        selected_agent_ids = self._get_checked_agents(
            self._ui.treeWidget.invisibleRootItem()
        )

        agent_ids_to_delete = set()

        if self._ui.checkBoxDeleteAgent.isChecked():
            for agent in selected_agent_ids:
                agent_ids_to_delete.add(agent)

        else:
            """deletes all contracts which are connected to selected agents"""
            # all contract which are connected to selected agents will be deleted
            for agent_id in selected_agent_ids:
                self._main_controller.delete_all_related_contracts(agent_id)

        if self._ui.checkBoxDeleteAllConnectedAgents.isChecked():
            agent_id_set = set(selected_agent_ids)
            for agent_id in selected_agent_ids:
                for (
                        related_agent_id
                ) in self._main_controller.scenario.get_all_related_agents_ids(
                    agent_id
                ):
                    agent_id_set.add(related_agent_id)

            if not self._ui.checkBoxDeleteAgent.isChecked():
                for agent_id in selected_agent_ids:
                    agent_id_set.remove(agent_id)

            for agent_id in agent_id_set:
                agent_ids_to_delete.add(agent_id)

        for agent_id_to_del in agent_ids_to_delete:
            self._main_controller.remove_agent(agent_id_to_del)

        self._ui.close()

    def cancel_button_clicked(self):
        """Closes the dialog"""
        self._ui.close()

    def init_ui(self):
        """Initializes the UI"""
        self._ui.treeWidget.setRootIsDecorated(False)
        self._ui.treeWidget.setColumnCount(3)
        self._ui.treeWidget.setHeaderLabels(
            [
                "Agent ID",
                "Agent type",
                "Amount of \nconnected Agents",
                "Amount of \ncontracts",
            ]
        )

        self._ui.treeWidget.setColumnWidth(0, 200)
        self._ui.treeWidget.setAlternatingRowColors(True)
        self._reset_attributes()
        self._ui.buttonBox.button(QDialogButtonBox.Ok).clicked.connect(
            self.confirm_button_clicked
        )
        self._ui.buttonBox.button(QDialogButtonBox.Cancel).clicked.connect(
            self.cancel_button_clicked
        )

    def _reset_attributes(self):
        """Resets all attributes to the default state"""
        self._ui.treeWidget.clear()
        self._tree_items.clear()

        for agent in self._agents:
            amount_of_related_contracts = self.scenario.get_amount_of_related_contracts(
                agent.id
            )
            amount_of_connected_agents = self.scenario.get_amount_connected_agents(
                agent.id
            )

            agent: AgentController
            item = AgentTreeItem(
                self._ui.treeWidget,
                agent.id,
                agent.type_name,
                amount_of_connected_agents,
                amount_of_related_contracts,
            )
            self._tree_items.append(item)



import logging
import typing
from collections import defaultdict
from typing import Optional, Dict, Any, List, Tuple

import yaml
from sqlalchemy import Column, Integer, String, inspect, MetaData, Table, text, Text, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from famegui.database.settings_db.init_db import init_db_session

Base = declarative_base()
_model_registry = {}  # Global registry to store model classes


def get_existing_table_names(yaml_path, session):
    with open(yaml_path, 'r') as file:
        yaml_content = yaml.safe_load(file)

    tables = yaml_content.get('db', {}).get('tables', {})
    inspector = inspect(session.bind)
    existing_tables = inspector.get_table_names()

    return existing_tables


def read_all_data(session):
    all_data = {}
    # Use the inspector to get the metadata of the database
    inspector = inspect(session.bind)

    all_data = {}

    # Get all table names from the database
    tables = inspector.get_table_names()

    # Loop over each table and retrieve its data
    for table_name in tables:
        query = session.execute(text(f"SELECT * FROM {table_name}"))
        rows = query.fetchall()

        # Get the column names from the table
        columns = query.keys()

        # Convert rows to dictionaries with column names as keys
        records = [dict(zip(columns, row)) for row in rows]

        # Store the result in all_data dictionary
        all_data[table_name] = records
    return all_data


def get_or_create_model(table_name, columns):
    """Get existing model class or create new one"""
    if table_name in _model_registry:
        return _model_registry[table_name]

    class_attrs = columns.copy()
    class_attrs['__tablename__'] = table_name
    ModelClass = type(table_name.capitalize(), (Base,), class_attrs)
    _model_registry[table_name] = ModelClass
    return ModelClass


def load_yaml_to_db(yaml_path, session):
    with open(yaml_path, 'r') as file:
        yaml_content = yaml.safe_load(file)

    tables = yaml_content.get('db', {}).get('tables', {})
    inspector = inspect(session.bind)
    existing_tables = inspector.get_table_names()

    for table_name, table_info in tables.items():
        table_schema = table_info.get('table_schema', {})
        preset_values = table_info.get('preset_values', [])

        # Build columns dynamically
        columns = {}
        columns['id'] = Column(Integer, primary_key=True, autoincrement=True)
        for column_name in table_schema.keys():
            if column_name == 'data_type':
                columns[column_name] = Column(Integer)
            else:
                columns[column_name] = Column(String)

        # Get or create model class
        ModelClass = get_or_create_model(table_name, columns)

        if table_name not in existing_tables:
            # Create the table only if it doesn't exist
            Base.metadata.create_all(session.bind, tables=[ModelClass.__table__])

            logging.info(f"Table '{table_name}' created.")

            insert_preset_values(ModelClass, preset_values, session, table_name)
        else:
            logging.info(f"Table '{table_name}' already exists. Skipping creation.")

            # avoid loading shortcut presets in
            # order to not override by user defined preferences

            insert_preset_values(ModelClass, preset_values, session, table_name)

    # Insert preset values


def insert_preset_values(ModelClass, preset_values, session, table_name):
    try:
        for new_record in preset_values:
            # Create filter conditions for all columns except 'id'
            filter_conditions = []
            for key, value in new_record.items():
                if key != 'id' and key != "shortcut_key":  # Exclude id from comparison
                    filter_conditions.append(getattr(ModelClass, key) == value)

            # Check if a record with the same values (excluding id) already exists
            existing_record = session.query(ModelClass).filter(*filter_conditions).first()

            if not existing_record:
                # Only insert if no matching record exists
                obj = ModelClass(**new_record)
                session.add(obj)
                logging.info(f"New preset record added to '{table_name}': {new_record}")
            else:
                logging.info(f"Skipping duplicate record in '{table_name}': {new_record}")

        session.commit()
        logging.info(f"Finished processing preset values for '{table_name}'.")
    except Exception as e:
        session.rollback()
        logging.info(f"Error inserting preset values into '{table_name}': {str(e)}")


def update_db(session, table_name, row_id, field_name, value,
              return_updated_record=False) -> typing.Union[None, dict]:
    try:
        # Reflect the table
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Check if the field exists in the table
        if field_name not in table.c:
            logging.warning(f"Field '{field_name}' does not exist in table '{table_name}'.")
            return

        # Build the update statement
        stmt = (
            table.update()
            .where(table.c.id == row_id)  # Assuming 'id' is the primary key column
            .values({field_name: value})
        )

        # Execute the update statement
        result = session.execute(stmt)

        # Check if any row was affected
        if result.rowcount == 0:
            logging.warning(f"No row with ID {row_id} found in table '{table_name}'.")
            return

        # Commit the transaction
        session.commit()
        logging.info(f"Updated field '{field_name}' to '{value}' in table '{table_name}', row ID {row_id}.")

        # Fetch the updated record if required
        if return_updated_record:
            select_stmt = select(table).where(table.c.id == row_id)
            result = session.execute(select_stmt)
            record = result.first()
            return dict(record._mapping) if record else None

        return None



    except SQLAlchemyError as e:
        # Rollback the transaction in case of any errors
        session.rollback()
        logging.error(f"Error updating field '{field_name}' in table '{table_name}': {str(e)}")


# Example usage:
# Assuming 'session' is your SQLAlchemy session object and 'yaml_path' is the path to your YAML file.
# load_yaml_to_db(yaml_path, session)


def get_rows(
        session: Session,
        table_name: str,
        fields_to_select: List[str],
        condition: Optional[Tuple[str, Any]] = None
) -> typing.Union[Optional[List[Dict[str, Any]]], None]:
    """
    Retrieve rows from a database table, with an optional filter condition, and return selected fields.

    Args:
        session (Session): The SQLAlchemy session object to interact with the database.
        table_name (str): The name of the table from which to retrieve the rows.
        fields_to_select (List[str]): A list of strings specifying which columns/fields to return from the table.
        condition (Optional[Tuple[str, Any]]): A tuple specifying the filter condition. The first element is the column
                                               name and the second element is the value to filter by. If no condition is
                                               provided, all rows will be retrieved.

    Returns:
        Optional[List[Dict[str, Any]]]: A list of dictionaries containing the selected fields and their values for
                                        each row. Returns None if no rows are found or if an error occurs.

    Raises:
        SQLAlchemyError: If any SQLAlchemy error occurs during the query execution.

    Example:
        # To get all rows
        results = get_rows(session, "users", ["id", "name", "email"])

        # To get rows with a condition
        results = get_rows(session, "users", ["id", "name", "email"], ("email", "example@example.com"))
    """
    try:
        # Reflect the table from the database
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Check if all the requested fields exist in the table
        for field in fields_to_select:
            if field not in table.c:
                logging.warning(f"Field '{field}' does not exist in table '{table_name}'.")
                return []

        # Build the select statement
        query = session.query(*[table.c[field] for field in fields_to_select])

        # Apply condition if provided
        if condition:
            filter_column, filter_value = condition
            if filter_column not in table.c:
                logging.warning(f"Field '{filter_column}' does not exist in table '{table_name}'.")
                return []
            query = query.filter(table.c[filter_column] == filter_value)

        # Execute the query and fetch all rows
        results = query.all()

        # If no results are found, return a message
        if not results:
            logging.warning(f"No matching records found in table '{table_name}'.")
            return []

        # Convert the results to a list of dictionaries
        rows_data = [{field: getattr(row, field) for field in fields_to_select} for row in results]

        return rows_data

    except SQLAlchemyError as e:
        # Handle any SQLAlchemy exceptions
        logging.error(f"Error retrieving data from table '{table_name}': {str(e)}")
        return None


def check_entry_exists(
        session: Session,
        table_name: str,
        filter_query_name: str,
        expected_value: Any
) -> bool:
    """
    Check if a specific entry exists in the database table based on a filter query.

    Args:
        session (Session): The SQLAlchemy session object to interact with the database.
        table_name (str): The name of the table to check the entry.
        filter_query_name (str): The column name used to filter rows (e.g., 'email' or 'id').
        expected_value (Any): The value to match in the filter (e.g., the specific email or id value).

    Returns:
        bool: True if the entry exists, False otherwise.

    Raises:
        SQLAlchemyError: If any SQLAlchemy error occurs during the query execution.
    """
    try:
        # Reflect the table from the database
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Build the select statement
        query = session.query(table).filter(table.c[filter_query_name] == expected_value)

        # Execute the query and check if any row exists
        result = query.first()

        # Return True if result is not None (entry exists), otherwise False
        return result is not None

    except SQLAlchemyError as e:
        # Handle any SQLAlchemy exceptions
        logging.error(f"Error checking if entry exists in table '{table_name}': {str(e)}")
        return False


def get_single_row(
        session: Session,
        table_name: str,
        filter_query_name: str,
        expected_value: Any,
        fields_to_select: List[str]
) -> Optional[Dict[str, Any]]:
    """
    Retrieve a single row from a database table based on a filter query and return selected fields.

    Args:
        session (Session): The SQLAlchemy session object to interact with the database.
        table_name (str): The name of the table from which to retrieve the row.
        filter_query_name (str): The column name used to filter rows (e.g., 'email' or 'id').
        expected_value (Any): The value to match in the filter (e.g., the specific email or id value).
        fields_to_select (List[str]): A list of strings specifying which columns/fields to return from the table.

    Returns:
        Optional[Dict[str, Any]]: A dictionary containing the selected fields and their values for the matching row.
        Returns None if no matching row is found or if an error occurs.

    Raises:
        SQLAlchemyError: If any SQLAlchemy error occurs during the query execution.

    Example:
        result = get_single_row(session, "users", "email", "example@example.com", ["id", "name", "email"])
    """

    try:
        # Reflect the table from the database
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Check if all the requested fields exist in the table
        for field in fields_to_select:
            if field not in table.c:
                logging.info(f"Field '{field}' does not exist in table '{table_name}'.")
                return None

        # Build the select statement
        query = session.query(*[table.c[field] for field in fields_to_select]) \
            .filter(table.c[filter_query_name] == expected_value)

        # Execute the query and fetch one row
        result = query.first()

        # If result is None, return a message
        if result is None:
            logging.warning(
                f"No matching record found in table '{table_name}' for {filter_query_name} = {expected_value}.")
            return None

        # Convert the result to a dictionary if the query succeeds
        row_data = {field: getattr(result, field) for field in fields_to_select}
        return row_data

    except SQLAlchemyError as e:
        # Handle any SQLAlchemy exceptions
        logging.error(f"Error retrieving data from table '{table_name}': {str(e)}")
        return None


def insert(session: Session, table_name: str, record: Dict[str, Any]) -> bool:
    """
    Insert a single row into a database table.

    Args:
        session (Session): The SQLAlchemy session object to interact with the database.
        table_name (str): The name of the table to insert the record into.
        record (Dict[str, Any]): A dictionary containing the column names and values to insert.

    Returns:
        bool: True if the insertion is successful, False otherwise.

    Raises:
        SQLAlchemyError: If any SQLAlchemy error occurs during the query execution.
    """

    try:
        # Reflect the table from the database
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Insert the record
        stmt = table.insert().values(**record)
        session.execute(stmt)
        session.commit()
        logging.info(f"Successfully inserted record into '{table_name}': {record}")
        return True

    except SQLAlchemyError as e:
        # Rollback in case of an error
        session.rollback()
        logging.error(f"Error inserting record into '{table_name}': {str(e)}")
        return False


def bulk_insert(session: Session, table_name: str, records: List[Dict[str, Any]]) -> bool:
    """
    Insert multiple rows into a database table in a bulk operation.

    Args:
        session (Session): The SQLAlchemy session object to interact with the database.
        table_name (str): The name of the table to insert the records into.
        records (List[Dict[str, Any]]): A list of dictionaries where each dictionary represents a row to insert.

    Returns:
        bool: True if the insertion is successful, False otherwise.

    Raises:
        SQLAlchemyError: If any SQLAlchemy error occurs during the query execution.
    """
    try:
        # Reflect the table from the database
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Insert the records in bulk
        stmt = table.insert().values(records)
        session.execute(stmt)
        session.commit()
        logging.info(f"Successfully inserted {len(records)} records into '{table_name}'")
        return True

    except SQLAlchemyError as e:
        # Rollback in case of an error
        session.rollback()
        logging.error(f"Error inserting records into '{table_name}': {str(e)}")
        return False


def get_row_count(session: Session, table_name: str) -> int:
    """
    Get the total number of rows in a database table.

    Args:
        session (Session): The SQLAlchemy session object to interact with the database.
        table_name (str): The name of the table to count rows from.

    Returns:
        int: The total number of rows in the table. Returns -1 if an error occurs.

    Raises:
        SQLAlchemyError: If any SQLAlchemy error occurs during the query execution.
    """
    try:
        # Reflect the table from the database
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=session.bind)

        # Get the row count
        count = session.query(table).count()
        logging.info(f"Table '{table_name}' has {count} rows.")
        return count

    except SQLAlchemyError as e:
        # Handle errors
        logging.error(f"Error getting row count from '{table_name}': {str(e)}")
        return -1


def get_string_set_values_from_db():
    """Fetch all String Set Groups with Values from project DB"""
    results = None

    with init_db_session() as session:
        results = get_rows(session, "db_string_item",
                           ["string_set_item_value", "parent_string_set_group_name"])

    results = results if results is not None else []

    # Group by `parent_string_set_group_name`
    grouped_data = defaultdict(list)

    for item in results:
        group_key = item["parent_string_set_group_name"]
        grouped_data[group_key].append(item["string_set_item_value"])

    # Convert defaultdict to a regular dict
    result = dict(grouped_data)


    return result


def get_string_set_values_for_group(group_name: str):
    results = None

    with init_db_session() as session:
        results = get_rows(session, "db_string_item",
                           ["string_set_item_value", "parent_string_set_group_name"])

    results = results if results is not None else []

    results = [
        item["string_set_item_value"] for item in results if
        item["parent_string_set_group_name"] == group_name
    ]

    return results


if __name__ == "__main__":
    pass



from famegui.models import Scenario


def get_contract_from_fields(scenario: Scenario, product_name: str,
                             receiver_id: int, sender_id: int):


    result_list = [contract for contract in scenario.contracts if contract.sender_id == sender_id
                   and contract.receiver_id == receiver_id and contract.product_name == product_name]

    if len(result_list) == 0:
        raise ValueError(f"Specified Contract does not exists:: check the fields "
                         f" product_name: {str(product_name)}; receiver_id: {str(receiver_id)}; sender_id: {str(receiver_id)} ")

    return result_list[0]



import logging
from datetime import datetime

from PySide6.QtCore import QTimer
from PySide6.QtGui import QFont, QTextCursor, Qt
from PySide6.QtWidgets import QTextEdit, QPushButton

from famegui.config.style_config import FAME_CONSOLE_TEXT_STYLE
from famegui.time_utils import WIDGET_DATE_FORMAT_CONSOLE


class GUIConsoleHandler(logging.StreamHandler):
    """A custom logging handler that writes to a QTextEdit to display the log messages in the GUI"""

    def __init__(self, text_edit: QTextEdit, clear_button: QPushButton):
        super().__init__()
        self._text_edit = text_edit
        self._text_edit.setReadOnly(True)

        font = QFont()
        font.setPointSize(10)
        self._text_edit.setFont(font)

        self._text_edit.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.TextSelectableByKeyboard)

        self._text_edit.setStyleSheet(FAME_CONSOLE_TEXT_STYLE)

        # Add a clear button
        clear_button.clicked.connect(self.clear_logs)

    def emit(self, record):
        """On new log message, append it to the QTextEdit and scroll to the bottom"""
        msg = self.format(record)
        current_time = datetime.now().strftime(WIDGET_DATE_FORMAT_CONSOLE)

        log_entry = f"{current_time}: {record.levelname}: {msg}\n"

        self._text_edit.moveCursor(QTextCursor.MoveOperation.End)
        self._text_edit.insertPlainText(log_entry)

        # Limit the number of lines to 150
        document = self._text_edit.document()
        if document.lineCount() > 150:
            cursor = QTextCursor(document)
            cursor.movePosition(QTextCursor.MoveOperation.Start)
            cursor.movePosition(QTextCursor.MoveOperation.Down, QTextCursor.MoveMode.KeepAnchor, document.lineCount() - 150)
            cursor.removeSelectedText()

        QTimer.singleShot(0, self.scroll_to_bottom)

    def scroll_to_bottom(self):
        """Scroll to the bottom of the text edit"""
        self._text_edit.verticalScrollBar().setValue(
            self._text_edit.verticalScrollBar().maximum()
        )

    def clear_logs(self):
        """Clear all log messages from the text edit"""
        self._text_edit.clear()



import os
import logging
import typing

from PySide6 import QtCore


class UserSettings:
    """Class to load and save user settings on the local host"""

    def __init__(self):
        self._settings = QtCore.QSettings("FAME", "famegui")

    _KEY_WORKING_DIR = "working_directory"

    @property
    def working_dir(self):
        dir_path = self._settings.value(self._KEY_WORKING_DIR)
        if dir_path is None:
            logging.warning("path to working dir is not defined")
        elif not os.path.isdir(dir_path):
            logging.warning("deleting invalid path to working dir: {}".format(dir_path))
            self._settings.remove(self._KEY_WORKING_DIR)
            dir_path = None
        return dir_path

    def set_working_dir(self, path: str):
        path = os.path.abspath(path)
        logging.info("configuring new working dir: {}".format(path))
        if not os.path.isdir(path):
            raise ValueError(
                "Can't use '{}' as new working directory: invalid path.".format(path)
            )
        return self._settings.setValue(self._KEY_WORKING_DIR, path)



import PySide6
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt, QPoint
from PySide6.QtGui import QCursor
from PySide6.QtWidgets import QGraphicsView


class MainGraphicsWidget(QGraphicsView):
    unselect_contracts = QtCore.Signal(name="unselect_contracts")
    agent_deletion_requested = QtCore.Signal(object)
    zoom_level_changed = QtCore.Signal(int)  # Signal to communicate zoom level changes

    def __init__(self, parent=None):
        super(MainGraphicsWidget, self).__init__(parent)
        self.startPos = None
        self._zoom_factor = 1.0  # Initialize zoom factor
        self.grabGesture(QtCore.Qt.GestureType.PanGesture)
        self.grabGesture(QtCore.Qt.GestureType.PinchGesture)

    def _get_local_widget_position(self) -> QPoint:
        """transforms the global mouse position to the local widget position"""
        mouse_position = QCursor.pos()
        local_position = self.mapFromGlobal(mouse_position)
        return local_position

    def keyPressEvent(self, event: PySide6.QtGui.QKeyEvent) -> None:
        """triggers agent deletion when delete key is pressed"""


        key = event.key()
        if key == Qt.Key.Key_Delete:
            self.agent_deletion_requested.emit(self._get_local_widget_position())
        super().keyPressEvent(event)

    def event(self, event):
        if event.type() == QtCore.QEvent.Type.Gesture:
            return self.gestureEvent(event)
        return super().event(event)

    def keyReleaseEvent(self, event):

        super().keyReleaseEvent(event)

    def get_agent_from_current_loc(self):
        """
        Retrieves the agent/item at the current mouse position in the scene.
        Returns:
            QGraphicsItem or None: The item at the current mouse position, or None if no item exists.
        """
        # Get global mouse position
        global_pos = QCursor.pos()

        # Convert to local widget position
        local_pos = self.mapFromGlobal(global_pos)

        # Convert to scene position
        scene_pos = self.mapToScene(local_pos)

        # Retrieve the item at the scene position
        if self.scene():
            item = self.scene().itemAt(scene_pos, self.transform())

            return item
        return None

    def gestureEvent(self, event):
        pan = event.gesture(QtCore.Qt.PanGesture)
        if pan:
            self.handlePanGesture(pan)
            return True
        pinch = event.gesture(QtCore.Qt.PinchGesture)
        if pinch:
            self.handlePinchGesture(pinch)
            return True
        return False

    def handlePanGesture(self, pan):
        delta = pan.delta()
        transform = self.transform()
        deltaX = delta.x() / transform.m11()
        deltaY = delta.y() / transform.m22()
        self.setSceneRect(self.sceneRect().translated(-deltaX, -deltaY))

    def handlePinchGesture(self, gesture):
        if gesture.changeFlags() & QtWidgets.QPinchGesture.ChangeFlag.ScaleFactorChanged:
            scale_factor = gesture.scaleFactor()
            self._zoom_factor *= scale_factor
            self._zoom_factor = max(0.000001, min(self._zoom_factor, 10))  # Clamp zoom factor
            self.resetTransform()
            self.scale(self._zoom_factor, self._zoom_factor)
            # Emit the zoom level changed signal
            zoom_level = int(self._zoom_factor * 1000)
            self.zoom_level_changed.emit(zoom_level)

    def mousePressEvent(self, event):
        """event handler for remounting the scene origin point int the canvas"""

        self.unselect_contracts.emit()
        if event.modifiers() & Qt.KeyboardModifier.ShiftModifier and event.button() == Qt.MouseButton.LeftButton:
            self.startPos = event.pos()


        else:
            super(MainGraphicsWidget, self).mousePressEvent(event)

    def wheelEvent(self, event: PySide6.QtGui.QWheelEvent) -> None:
        """prevent from unwanted movement/ side effects when scrolling with"""
        if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:  # bitwise comparison necessary, due to the fact that multiple
            # modifiers can be pressed at the same time
            return
        super().wheelEvent(event)

    def mouseMoveEvent(self, event):
        """compute the difference between the current cursor position and the
        previous saved origin point"""
        if self.startPos is not None:
            delta = self.startPos - event.pos()
            transform = self.transform()
            deltaX = delta.x() / transform.m11()
            deltaY = delta.y() / transform.m22()
            self.setSceneRect(self.sceneRect().translated(deltaX, deltaY))
            self.startPos = event.pos()
        else:
            super(MainGraphicsWidget, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """resets the alternative origin point"""
        self.startPos = None
        super().mouseReleaseEvent(event)



from typing import List

import PySide6
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QPointF
from PySide6.QtGui import Qt
from PySide6.QtWidgets import QGraphicsSceneMouseEvent

from famegui.agent_controller import AgentController
from famegui.scenario_canvas_elements.agent_canvas_item import AgentGraphItem
from famegui.scenario_canvas_elements.contract_canvas_item import ContractGraphItem


class ScenarioGraphView(QtWidgets.QGraphicsScene):
    """View displaying the scenario as a graph"""

    # (x, y)
    agent_creation_requested = QtCore.Signal(int, int)
    # (agent_id)
    agent_edition_requested = QtCore.Signal(int)
    agent_multi_edition = QtCore.Signal(list)

    # (sender_id, receiver_id)
    contract_creation_requested = QtCore.Signal(int, int)
    # agent_id (can be None)
    selected_agent_changed = QtCore.Signal(int)
    # zoom factor control
    zoom_in_requested = QtCore.Signal()
    move_view_left_requested = QtCore.Signal()
    move_view_right_requested = QtCore.Signal()

    # agent_id, old_x, old_y, new_x, new_y
    agent_dragged = QtCore.Signal()

    # multi selection
    released_multi_selection_mode = QtCore.Signal(list, list)
    released_multi_selection_mode_no_valid = QtCore.Signal()

    zoom_out_requested = QtCore.Signal()
    multi_selection_mode_released = QtCore.Signal()

    def __init__(self,
                 on_selected_agent_released,
                 parent=None):
        QtWidgets.QGraphicsScene.__init__(self, parent)
        self.selectionChanged.connect(self._on_scene_selection_changed)
        self._selected_source_agents = []
        self._selected_agents_to_edit = []
        self._selected_target_agents = []
        self._is_in_multi_selection_mode = False
        self.last_state_alt_pressed = False
        self._on_selected_agent_released = on_selected_agent_released

        self._original_pos: QPointF = None
        self._original_pos_related_agent_id = -1

    def is_in_multi_selection_mode(self) -> bool:
        '''Return True if the multi agent selection mode is enabled'''
        return self._is_in_multi_selection_mode

    def set_multi_selection_mode(self, in_selection_mode: bool):
        '''Enable or disable multi selection mode using a flag'''
        self._is_in_multi_selection_mode = in_selection_mode

    def clearSelection(self, clear_highlights=False) -> None:
        if clear_highlights:
            for agent in self.get_agent_items():
                agent.set_border_highlight(False)
                agent.setSelected(False)

        super().clearSelection()

    @property
    def selected_agent_id(self):
        items = self.selectedItems()
        if len(items) == 1:
            item = items[0]
            assert item.type() == AgentGraphItem.Type
            return item.agent_id
        return None

    def remove_contract(self, agent_sender_id: int, agent_receiver_id: int):
        '''Remove a contract between two agents in the graph '''
        to_delete = []

        for contract in self.get_contract_items():
            contract: ContractGraphItem

            if contract.sourceNode().agent_id == agent_sender_id and contract.destNode().agent_id == agent_receiver_id:
                to_delete.append(contract)
                break

        for contract_to_delete in to_delete:
            self.removeItem(contract_to_delete)

    def get_contract_items(self) -> List[ContractGraphItem]:
        ''' get all contract items in the graph '''
        return_list = []
        for item in self.items():
            if issubclass(type(item), ContractGraphItem):
                return_list.append(item)
        return return_list

    def update_all_contracts(self):
        ''' Adjust all Contract Positions '''
        return_list = []
        for item in self.items():
            if issubclass(type(item), ContractGraphItem):
                return_list.append(item)

        for c in return_list:
            c: ContractGraphItem
            c.adjust()

    def get_agent_items(self) -> List[AgentGraphItem]:
        ''' get all agent items in the graph '''

        return_list = []
        for item in self.items():
            if issubclass(type(item), AgentGraphItem):
                return_list.append(item)
        return return_list

    def remove_agents(self, agent_ids: List[int]):
        '''Remove a list of agents from the graph'''

        for agent_id in agent_ids:
            self.remove_agent(agent_id)

    def remove_related_contracts(self, agent_id: int):
        '''Remove all contracts related to an agent from the graph'''
        to_delete = []
        for contract in self.get_contract_items():
            contract: ContractGraphItem
            if contract.sourceNode().agent_id == agent_id or contract.destNode().agent_id == agent_id:
                to_delete.append(contract)

        for idx, contract in enumerate(to_delete):
            contract.prepareGeometryChange()
            for child_item in contract.childItems():
                super().removeItem(child_item)

            self.removeItem(contract)

    def blockSignals(self, b):

        return super().blockSignals(b)

    def remove_agent(self, agent_id: int):
        '''Permanently remove an agent from the graph'''
        agents_to_delete = []
        for agent in self.get_agent_items():
            if agent.agent_id == agent_id:
                agents_to_delete.append(agent)

        self.remove_related_contracts(agent_id)

        for agent in agents_to_delete:
            self.removeItem(agent)

    def add_agent(self, agent_ctrl: AgentController):
        item = AgentGraphItem(agent_ctrl.id, agent_ctrl.type_name, agent_ctrl.svg_color)
        item.setToolTip(agent_ctrl.tooltip_text)
        item.setPos(agent_ctrl.x, agent_ctrl.y)
        item.setZValue(100.0)
        self.addItem(item)
        agent_ctrl.set_scene_item(item)

    def add_contract(self, sender: AgentController, receiver: AgentController):
        self.addItem(ContractGraphItem(sender.scene_item, receiver.scene_item))

    def event(self, event: PySide6.QtCore.QEvent) -> bool:
        '''Handles user events of the scene'''

        if issubclass(type(event), QGraphicsSceneMouseEvent):

            if ((event.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier)
                    and (event.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier)):

                click_pos = event.scenePos()

                if event.button() == QtCore.Qt.MouseButton.LeftButton:
                    for item in self.items(click_pos):
                        if item.type() == AgentGraphItem.Type:
                            item: AgentGraphItem
                            item.set_selected_for_edit(True)

            event: QGraphicsSceneMouseEvent
            if event.modifiers() == QtCore.Qt.KeyboardModifier.ShiftModifier:
                self.set_multi_selection_mode(True)
                click_pos = event.scenePos()

                if event.button() == QtCore.Qt.MouseButton.LeftButton:
                    for item in self.items(click_pos):
                        if item.type() == AgentGraphItem.Type:
                            item: AgentGraphItem
                            item.setSelected(True)
                            if not self._selected_source_agents.__contains__(item):
                                self._selected_source_agents.append(item)

                if event.button() == QtCore.Qt.MouseButton.RightButton:
                    for item in self.items(click_pos):
                        if item.type() == AgentGraphItem.Type:
                            item: AgentGraphItem
                            item.setSelected(True, False)
                            if not self._selected_target_agents.__contains__(item):
                                self._selected_target_agents.append(item)

        return super().event(event)

    def mouseReleaseEvent(self, event):
        """
        Handles the mouse release event to capture the release position
        and notify the on_selected_agent_released callback with the start
        and end positions.
        """
        click_pos = event.scenePos()

        # Check if an agent was clicked and moved
        for item in self.items(click_pos):
            if item.type() == AgentGraphItem.Type:
                item: AgentGraphItem

                # Capture the release (end) position
                end_x = click_pos.x()
                end_y = click_pos.y()

                # Pass the start and end positions to the callback
                if self._original_pos is not None and item.start_pos is not None:
                    start_x = item.start_pos.x()
                    start_y = item.start_pos.y()
                    self._on_selected_agent_released(item.agent_id, start_x, start_y, end_x, end_y)

                # Reset the original position after release
                self._original_pos = None
                self._original_pos_related_agent_id = -1

        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        if event.button() != QtCore.Qt.MouseButton.LeftButton:
            return
        click_pos = event.scenePos()
        # is the double click on an agent item?
        for item in self.items(click_pos):
            if item.type() == AgentGraphItem.Type:
                self.agent_edition_requested.emit(item.agent_id)
                return

        self.agent_creation_requested.emit(click_pos.x(), click_pos.y())

    def keyReleaseEvent(self, event: PySide6.QtGui.QKeyEvent) -> None:
        '''Handles user key events of the scene'''
        '''Usually used to exit multi selection mode'''

        for item in self._selected_target_agents:
            item.setSelected(False)
        for item in self._selected_source_agents:
            item.setSelected(False)

        selected_agents_for_multi_edit = []

        for item in self.get_agent_items():

            if item.is_selected_for_edit():
                selected_agents_for_multi_edit.append(item.agent_id)

            item.set_selected_for_edit(False)

        if len(selected_agents_for_multi_edit) > 0:
            self.agent_multi_edition.emit(selected_agents_for_multi_edit)

        if event.key() == Qt.Key.Key_Shift:
            self.set_multi_selection_mode(False)
            # EXIT MULTI SELECTION MODE
            if len(self._selected_source_agents) != 0 and len(self._selected_target_agents) != 0:
                self.released_multi_selection_mode.emit(
                    [x.agent_id for x in self._selected_target_agents],
                    [x.agent_id
                     for x in self._selected_source_agents])
                self._selected_target_agents.clear()
                self._selected_source_agents.clear()
                for item in self.get_contract_items():
                    item.setSelected(False)
                super().keyReleaseEvent(event)
                return
            self._selected_target_agents.clear()
            self._selected_source_agents.clear()
            self.released_multi_selection_mode_no_valid.emit()
        super().keyReleaseEvent(event)

    def get_agent_id(self, q_point: QPointF):
        '''Helper to return the agent id of the agent at the given position'''
        for item in self.items(q_point):
            if item.type() == AgentGraphItem.Type:
                return item.agent_id

    def _get_agent_id(self, event):
        '''Returns the agent id of the agent at the given position'''
        click_pos = event.scenePos()
        for item in self.items(click_pos):
            if item.type() == AgentGraphItem.Type:
                return item.agent_id

    def mousePressEvent(self, event):
        '''Handles user mouse events of the scene'''
        # Multi SELECT MODE

        self.views()[0].viewport().setCursor(Qt.CursorShape.ClosedHandCursor)

        click_pos = event.scenePos()

        for item in self.items(click_pos):
            if item.type() == AgentGraphItem.Type:
                self._original_pos = click_pos


                self._original_pos_related_agent_id = item.agent_id

        if event.button() == QtCore.Qt.MouseButton.RightButton and event.modifiers() != QtCore.Qt.KeyboardModifier.ShiftModifier:
            for item in self.items(click_pos):
                if item.type() == AgentGraphItem.Type:
                    item: AgentGraphItem
                    self._on_agent_right_clicked(item.agent_id)
                    return
        QtWidgets.QGraphicsScene.mousePressEvent(self, event)

    def wheelEvent(self, event):
        '''enable zoom in/out if ctrl key is pressed'''
        if event.modifiers() == QtCore.Qt.KeyboardModifier.ShiftModifier:
            if event.delta() > 0:
                self.move_view_left_requested.emit()
            else:
                self.move_view_right_requested.emit()
            event.accept()

        if QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.KeyboardModifier.ControlModifier:
            if event.delta() > 0:
                self.zoom_in_requested.emit()
            else:
                self.zoom_out_requested.emit()
            event.accept()

    def _on_agent_right_clicked(self, agent_id):
        source_id = self.selected_agent_id
        if source_id is not None and source_id != agent_id:
            self.contract_creation_requested.emit(source_id, agent_id)

    def _on_scene_selection_changed(self):
        self.selected_agent_changed.emit(self.selected_agent_id)



def remove_none_values_and_empty_elements(item):
    """Recursively remove all None values and empty elements from a dictionary or list to clean up the data
    prepared for saving to a file
    """
    if isinstance(item, dict):
        # Handle dictionary: recursively clean its key-value pairs
        new_dict = {
            k: remove_none_values_and_empty_elements(v)
            for k, v in item.items()
            if v is not None
        }

        return {k: v for k, v in new_dict.items() if v not in [None, {}, []]}
    elif isinstance(item, list):
        # Handle list: recursively clean its elements
        new_list = [remove_none_values_and_empty_elements(v) for v in item]
        return [v for v in new_list if v not in [None, {}, []]]
    else:
        # Base case: return the item as-is
        return item


import json
import timeit
from typing import Dict, Any, Tuple, Optional


class DictDiffer:
    """
        A utility class for comparing two dictionaries and identifying differences.
    """

    @staticmethod
    def diff_direct(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> dict[str, dict[str, Optional[Any]]]:
        """
        Compares two dictionaries and identifies differences in their keys and values.

        This method uses sets to efficiently find differences and a dictionary comprehension
        to build the result.

        Args:
             dict1 (Dict[str, Any]): The first dictionary to compare.
             dict2 (Dict[str, Any]): The second dictionary to compare.

        Returns:
                Dict[str, Dict[str, Any]]: A dictionary containing the keys that are different
                                            between `dict1` and `dict2`, along with their respective
                                            old and new values.
                                            - If a key is missing in one of the dictionaries,
                                            the corresponding value will be `None`.
                                            Example:
                                                {
                                                    "key1": {"old_value": 1, "new_value": 2},
                                                    "key2": {"old_value": None, "new_value": 3}
                                                }
        """
        keys1, keys2 = set(dict1), set(dict2)

        # Find keys that are different
        diff_keys = {
            k: {
                "old_value": dict1.get(k),
                "new_value": dict2.get(k)
            }
            for k in keys1 | keys2
            if (k not in dict1) or (k not in dict2) or dict1[k] != dict2[k]
        }
        return diff_keys
